import base64
import binascii

from pcs.common import reports
from pcs.common.node_communicator import (
    Request,
    RequestData,
)
from pcs.common.reports.item import ReportItem
from pcs.lib.communication.tools import (
    AllAtOnceStrategyMixin,
    AllSameDataMixin,
    RunRemotelyBase,
    SkipOfflineMixin,
    SimpleResponseProcessingMixin,
    SimpleResponseProcessingNoResponseOnSuccessMixin,
)
from pcs.lib.communication.qdevice import QdeviceBase


class GetCaCert(AllSameDataMixin, AllAtOnceStrategyMixin, RunRemotelyBase):
    def __init__(self, report_processor):
        super().__init__(report_processor)
        self._data = []

    def _get_request_data(self):
        return RequestData("remote/qdevice_net_get_ca_certificate")

    def _process_response(self, response):
        report = self._get_response_report(response)
        if report is not None:
            self._report(report)
            return
        target = response.request.target
        try:
            self._data.append((target, base64.b64decode(response.data)))
        except (TypeError, binascii.Error):
            self._report(
                ReportItem.error(
                    reports.messages.InvalidResponseFormat(target.label)
                )
            )

    def on_complete(self):
        return self._data


class ClientSetup(
    SimpleResponseProcessingNoResponseOnSuccessMixin,
    SkipOfflineMixin,
    AllSameDataMixin,
    AllAtOnceStrategyMixin,
    RunRemotelyBase,
):
    def __init__(
        self,
        report_processor,
        ca_cert,
        skip_offline_targets=False,
        allow_skip_offline=True,
    ):
        super().__init__(report_processor)
        if allow_skip_offline:
            self._set_skip_offline(skip_offline_targets)
        self._ca_cert = ca_cert

    def _get_request_data(self):
        return RequestData(
            "remote/qdevice_net_client_init_certificate_storage",
            [("ca_certificate", base64.b64encode(self._ca_cert))],
        )


class SignCertificate(AllAtOnceStrategyMixin, RunRemotelyBase):
    def __init__(self, report_processor):
        super().__init__(report_processor)
        self._output_data = []
        self._input_data = []

    def add_request(self, target, cert, cluster_name):
        self._input_data.append((target, cert, cluster_name))

    def _prepare_initial_requests(self):
        return [
            Request(
                target,
                RequestData(
                    "remote/qdevice_net_sign_node_certificate",
                    [
                        ("certificate_request", base64.b64encode(cert)),
                        ("cluster_name", cluster_name),
                    ],
                ),
            )
            for target, cert, cluster_name in self._input_data
        ]

    def _process_response(self, response):
        report = self._get_response_report(response)
        if report is not None:
            self._report(report)
            return
        target = response.request.target
        try:
            self._output_data.append((target, base64.b64decode(response.data)))
        except (TypeError, binascii.Error):
            self._report(
                ReportItem.error(
                    reports.messages.InvalidResponseFormat(target.label)
                )
            )

    def on_complete(self):
        return self._output_data


class ClientImportCertificateAndKey(
    SimpleResponseProcessingMixin,
    SkipOfflineMixin,
    AllSameDataMixin,
    AllAtOnceStrategyMixin,
    RunRemotelyBase,
):
    def __init__(
        self,
        report_processor,
        pk12,
        skip_offline_targets=False,
        allow_skip_offline=True,
    ):
        super().__init__(report_processor)
        if allow_skip_offline:
            self._set_skip_offline(skip_offline_targets)
        self._pk12 = pk12

    def _get_request_data(self):
        return RequestData(
            "remote/qdevice_net_client_import_certificate",
            [("certificate", base64.b64encode(self._pk12))],
        )

    def _get_success_report(self, node_label):
        return ReportItem.info(
            reports.messages.QdeviceCertificateAcceptedByNode(node_label)
        )


class ClientDestroy(SimpleResponseProcessingMixin, QdeviceBase):
    def _get_request_data(self):
        return RequestData("remote/qdevice_net_client_destroy")

    def _get_success_report(self, node_label):
        return ReportItem.info(
            reports.messages.QdeviceCertificateRemovedFromNode(node_label)
        )



# pylint: disable=too-many-lines
import json

from textwrap import dedent
from unittest import mock, TestCase

from pcs_test.tools.misc import (
    dict_to_modifiers,
    get_tmp_file,
)

from pcs import cluster
from pcs.cli.common.errors import CmdLineInputError
from pcs.common.corosync_conf import (
    CorosyncConfDto,
    CorosyncNodeAddressDto,
    CorosyncNodeDto,
    CorosyncQuorumDeviceSettingsDto,
)
from pcs.common.interface import dto
from pcs.common.reports import codes as report_codes
from pcs.common.types import CorosyncTransportType


def _node(name, **kwargs):
    """node dictionary fixture"""
    return dict(name=name, **kwargs)


DEFAULT_TRANSPORT_TYPE = "knet"


class ClusterSetup(TestCase):
    # pylint: disable=too-many-public-methods
    def setUp(self):
        self.lib = mock.Mock(spec_set=["cluster"])
        self.cluster = mock.Mock(spec_set=["setup", "setup_local"])
        self.lib.cluster = self.cluster
        self.cluster_name = "cluster_name"

    @staticmethod
    def _get_default_kwargs():
        return dict(
            transport_type=None,
            transport_options={},
            link_list=[],
            compression_options={},
            crypto_options={},
            totem_options={},
            quorum_options={},
            force_flags=[],
        )

    def assert_setup_called_with(self, node_list, **kwargs):
        default_kwargs = self._get_default_kwargs()
        default_kwargs.update(
            dict(wait=False, start=False, enable=False, no_keys_sync=False)
        )
        default_kwargs.update(kwargs)
        self.cluster.setup.assert_called_once_with(
            self.cluster_name, node_list, **default_kwargs
        )

    def assert_setup_local_called_with(self, node_list, **kwargs):
        default_kwargs = self._get_default_kwargs()
        default_kwargs.update(kwargs)
        self.cluster.setup_local.assert_called_once_with(
            self.cluster_name, node_list, **default_kwargs
        )

    def call_cmd_without_cluster_name(self, argv, modifiers=None):
        cluster.cluster_setup(
            self.lib, argv, dict_to_modifiers(modifiers or {})
        )

    def call_cmd(self, argv, modifiers=None):
        self.call_cmd_without_cluster_name(
            [self.cluster_name] + argv, modifiers
        )

    def test_no_cluster_name(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd_without_cluster_name([])
        self.assertIsNone(cm.exception.message)

    def test_no_node(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd([])
        self.assertIsNone(cm.exception.message)

    def test_invalid_node_name(self):
        node = "node="
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd([node])
        self.assertEqual(
            "Invalid character '=' in node name '{}'".format(node),
            cm.exception.message,
        )

    def test_node_defined_multiple_times(self):
        node = "node"
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd([node, node])
        self.assertEqual(
            "Node name '{}' defined multiple times".format(node),
            cm.exception.message,
        )

    def test_one_node_no_addrs(self):
        node_name = "node"
        self.call_cmd([node_name])
        self.assert_setup_called_with([_node(node_name)])

    def test_one_node_empty_addrs(self):
        node_name = "node"
        self.call_cmd([node_name, "addr="])
        self.assert_setup_called_with([_node(node_name, addrs=[""])])

    def test_one_node_with_single_address(self):
        node_name = "node"
        addr = "node_addr"
        self.call_cmd([node_name, "addr={}".format(addr)])
        self.assert_setup_called_with([_node(node_name, addrs=[addr])])

    def test_one_node_with_multiple_addresses(self):
        node_name = "node"
        addr_list = ["addr{}".format(i) for i in range(3)]
        self.call_cmd([node_name] + [f"addr={addr}" for addr in addr_list])
        self.assert_setup_called_with([_node(node_name, addrs=addr_list)])

    def test_node_unknown_options(self):
        node_name = "node"
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(
                [node_name, "unknown=option", "another=one", "addr=addr"]
            )
        self.assertEqual(
            "Unknown options 'another', 'unknown' for node 'node'",
            cm.exception.message,
        )

    def test_multiple_nodes(self):
        self.call_cmd(
            ["node1", "addr=addr1", "addr=addr2", "node2", "node3", "addr=addr"]
        )
        self.assert_setup_called_with(
            [
                _node("node1", addrs=["addr1", "addr2"]),
                _node("node2"),
                _node("node3", addrs=["addr"]),
            ]
        )

    def test_multiple_nodes_without_addrs(self):
        node_list = ["node{}".format(i) for i in range(4)]
        self.call_cmd(node_list)
        self.assert_setup_called_with([_node(node) for node in node_list])

    def test_transport_type_missing(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(["node", "transport"])
        self.assertEqual("Transport type not defined", cm.exception.message)

    def test_transport_type_unknown(self):
        node = "node"
        self.call_cmd(
            [
                node,
                "transport",
                "unknown",
                "a=1",
                "link",
                "b=2",
                "c=3",
            ]
        )
        self.assert_setup_called_with(
            [_node(node)],
            transport_type="unknown",
            transport_options=dict(a="1"),
            link_list=[
                dict(b="2", c="3"),
            ],
        )

    def test_transport_with_unknown_keywords(self):
        node = "node"
        self.call_cmd(["node", "transport", "udp", "crypto", "a=1"])
        self.assert_setup_called_with(
            [_node(node)],
            transport_type="udp",
            crypto_options=dict(a="1"),
        )

    def test_transport_independent_options(self):
        node = "node"
        self.call_cmd([node, "quorum", "a=1", "b=2", "totem", "c=3", "a=2"])
        self.assert_setup_called_with(
            [_node(node)],
            quorum_options=dict(a="1", b="2"),
            totem_options=dict(c="3", a="2"),
        )

    def test_knet_links(self):
        node = "node"
        self.call_cmd(
            [
                node,
                "transport",
                "knet",
                "link",
                "c=3",
                "a=2",
                "link",
                "a=1",
            ]
        )
        self.assert_setup_called_with(
            [_node(node)],
            transport_type="knet",
            link_list=[
                dict(c="3", a="2"),
                dict(a="1"),
            ],
        )

    def test_knet_links_repetable_correctly(self):
        node = "node"
        self.call_cmd(
            [
                node,
                "transport",
                "knet",
                "link",
                "c=3",
                "a=2",
                "compression",
                "d=1",
                "e=1",
                "link",
                "a=1",
                "compression",
                "f=1",
            ]
        )
        self.assert_setup_called_with(
            [_node(node)],
            transport_type="knet",
            link_list=[
                dict(c="3", a="2"),
                dict(a="1"),
            ],
            compression_options=dict(d="1", e="1", f="1"),
        )

    def test_transport_not_repeatable(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(
                ["node", "transport", "knet", "a=1", "transport", "knet"]
            )
        self.assertEqual(
            "'transport' cannot be used more than once", cm.exception.message
        )

    def test_totem_not_repeatable(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(["node", "totem", "a=1", "b=1", "totem", "c=2"])
        self.assertEqual(
            "'totem' cannot be used more than once", cm.exception.message
        )

    def test_quorum_not_repeatable(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(["node", "quorum", "a=1", "b=1", "quorum", "c=2"])
        self.assertEqual(
            "'quorum' cannot be used more than once", cm.exception.message
        )

    def test_full_knet(self):
        self.call_cmd(
            [
                "node0",
                "node2",
                "addr=addr0",
                "node1",
                "addr=addr1",
                "addr=addr2",
                "totem",
                "a=1",
                "b=1",
                "quorum",
                "c=1",
                "d=1",
                "transport",
                "knet",
                "a=a",
                "b=b",
                "compression",
                "a=1",
                "b=2",
                "c=3",
                "crypto",
                "d=4",
                "e=5",
                "link",
                "aa=1",
                "link",
                "ba=1",
                "bb=2",
                "link",
                "ca=1",
                "cb=2",
                "cc=3",
            ]
        )
        self.assert_setup_called_with(
            [
                _node("node0"),
                _node("node2", addrs=["addr0"]),
                _node("node1", addrs=["addr1", "addr2"]),
            ],
            totem_options=dict(a="1", b="1"),
            quorum_options=dict(c="1", d="1"),
            transport_type="knet",
            transport_options=dict(a="a", b="b"),
            compression_options=dict(a="1", b="2", c="3"),
            crypto_options=dict(d="4", e="5"),
            link_list=[
                dict(aa="1"),
                dict(ba="1", bb="2"),
                dict(ca="1", cb="2", cc="3"),
            ],
        )

    def assert_with_all_options(self, transport_type):
        self.call_cmd(
            [
                "node0",
                "node2",
                "addr=addr0",
                "node1",
                "addr=addr1",
                "addr=addr2",
                "totem",
                "a=1",
                "b=1",
                "quorum",
                "c=1",
                "d=1",
                "transport",
                transport_type,
                "a=a",
                "b=b",
                "link",
                "aa=1",
                "link",
                "ba=1",
                "bb=2",
            ]
        )
        self.assert_setup_called_with(
            [
                _node("node0"),
                _node("node2", addrs=["addr0"]),
                _node("node1", addrs=["addr1", "addr2"]),
            ],
            totem_options=dict(a="1", b="1"),
            quorum_options=dict(c="1", d="1"),
            transport_type=transport_type,
            transport_options=dict(a="a", b="b"),
            link_list=[
                dict(aa="1"),
                dict(ba="1", bb="2"),
            ],
        )

    def test_full_udp(self):
        self.assert_with_all_options("udp")

    def test_full_udpu(self):
        self.assert_with_all_options("udpu")

    def test_enable(self):
        node_name = "node"
        self.call_cmd([node_name], {"enable": True})
        self.assert_setup_called_with([_node(node_name)], enable=True)

    def test_start(self):
        node_name = "node"
        self.call_cmd([node_name], {"start": True})
        self.assert_setup_called_with([_node(node_name)], start=True)

    def test_enable_start(self):
        node_name = "node"
        self.call_cmd([node_name], {"enable": True, "start": True})
        self.assert_setup_called_with(
            [_node(node_name)], enable=True, start=True
        )

    def test_wait(self):
        node_name = "node"
        self.call_cmd([node_name], {"wait": "10"})
        self.assert_setup_called_with([_node(node_name)], wait="10")

    def test_start_wait(self):
        node_name = "node"
        self.call_cmd([node_name], {"start": True, "wait": None})
        self.assert_setup_called_with([_node(node_name)], start=True, wait=None)

    def test_start_wait_timeout(self):
        node_name = "node"
        self.call_cmd([node_name], {"start": True, "wait": "10"})
        self.assert_setup_called_with([_node(node_name)], start=True, wait="10")

    def test_force(self):
        node_name = "node"
        self.call_cmd([node_name], {"force": True})
        self.assert_setup_called_with(
            [_node(node_name)],
            force_flags=[report_codes.FORCE],
        )

    def test_all_modifiers(self):
        node_name = "node"
        self.call_cmd(
            [node_name],
            {
                "force": True,
                "enable": True,
                "start": True,
                "wait": "15",
                "no-keys-sync": True,
            },
        )
        self.assert_setup_called_with(
            [_node(node_name)],
            enable=True,
            start=True,
            wait="15",
            no_keys_sync=True,
            force_flags=[report_codes.FORCE],
        )

    def test_live_with_local_modifiers(self):
        node_name = "node"
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd([node_name], {"overwrite": True})
        self.assertEqual(
            "Cannot specify '--overwrite' when '--corosync_conf' is not "
            "specified",
            cm.exception.message,
        )

    def test_corosync_conf_not_supported_modifiers(self):
        node_name = "node"
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(
                [node_name],
                {
                    "enable": True,
                    "start": True,
                    "wait": "15",
                    "no-keys-sync": True,
                    "corosync_conf": "file_path",
                    "force": True,
                },
            )
        self.assertEqual(
            "Cannot specify any of '--enable', '--no-keys-sync', '--start', "
            "'--wait' when '--corosync_conf' is specified",
            cm.exception.message,
        )

    def test_corosync_conf(self):
        node_name = "node"
        corosync_conf_data = b"new corosync.conf"
        self.cluster.setup_local.return_value = corosync_conf_data
        with get_tmp_file("test_cluster_corosync.conf", "rb") as output_file:
            self.call_cmd(
                [node_name],
                {"corosync_conf": output_file.name, "overwrite": True},
            )
            self.assertEqual(output_file.read(), corosync_conf_data)

        self.assert_setup_local_called_with([_node(node_name)])

    def test_corosync_conf_full_knet(self):
        corosync_conf_data = b"new corosync.conf"
        self.cluster.setup_local.return_value = corosync_conf_data
        with get_tmp_file("test_cluster_corosync.conf", "rb") as output_file:
            self.call_cmd(
                [
                    "node0",
                    "node2",
                    "addr=addr0",
                    "node1",
                    "addr=addr1",
                    "addr=addr2",
                    "totem",
                    "a=1",
                    "b=1",
                    "quorum",
                    "c=1",
                    "d=1",
                    "transport",
                    "knet",
                    "a=a",
                    "b=b",
                    "compression",
                    "a=1",
                    "b=2",
                    "c=3",
                    "crypto",
                    "d=4",
                    "e=5",
                    "link",
                    "aa=1",
                    "link",
                    "ba=1",
                    "bb=2",
                    "link",
                    "ca=1",
                    "cb=2",
                    "cc=3",
                ],
                {"corosync_conf": output_file.name, "overwrite": True},
            )
            self.assertEqual(output_file.read(), corosync_conf_data)
        self.assert_setup_local_called_with(
            [
                _node("node0"),
                _node("node2", addrs=["addr0"]),
                _node("node1", addrs=["addr1", "addr2"]),
            ],
            totem_options=dict(a="1", b="1"),
            quorum_options=dict(c="1", d="1"),
            transport_type="knet",
            transport_options=dict(a="a", b="b"),
            compression_options=dict(a="1", b="2", c="3"),
            crypto_options=dict(d="4", e="5"),
            link_list=[
                dict(aa="1"),
                dict(ba="1", bb="2"),
                dict(ca="1", cb="2", cc="3"),
            ],
        )


class NodeAdd(TestCase):
    # pylint: disable=too-many-public-methods
    def setUp(self):
        self.lib = mock.Mock(spec_set=["cluster"])
        self.cluster = mock.Mock(spec_set=["add_nodes"])
        self.lib.cluster = self.cluster
        self.hostname = "hostname"
        self._default_kwargs = dict(
            wait=False,
            start=False,
            enable=False,
            no_watchdog_validation=False,
            force_flags=[],
        )

    def assert_called_with(self, node_list, **kwargs):
        default_kwargs = dict(self._default_kwargs)
        default_kwargs.update(kwargs)
        self.cluster.add_nodes.assert_called_once_with(
            nodes=node_list, **default_kwargs
        )

    def call_cmd(self, argv, modifiers=None):
        cluster.node_add(self.lib, argv, dict_to_modifiers(modifiers or {}))

    def test_no_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd([])
        self.assertIsNone(cm.exception.message)

    def test_minimal(self):
        self.call_cmd([self.hostname])
        self.assert_called_with([_node(self.hostname)])

    def test_with_addr(self):
        addr = "addr1"
        self.call_cmd([self.hostname, f"addr={addr}"])
        self.assert_called_with([_node(self.hostname, addrs=[addr])])

    def test_with_empty_addr(self):
        self.call_cmd([self.hostname, "addr="])
        self.assert_called_with([_node(self.hostname, addrs=[""])])

    def test_with_multiple_addrs(self):
        addr_list = [f"addr{i}" for i in range(5)]
        self.call_cmd([self.hostname] + [f"addr={addr}" for addr in addr_list])
        self.assert_called_with([_node(self.hostname, addrs=addr_list)])

    def test_with_watchdog(self):
        watchdog = "watchdog_path"
        self.call_cmd([self.hostname, f"watchdog={watchdog}"])
        self.assert_called_with([_node(self.hostname, watchdog=watchdog)])

    def test_with_empty_watchdog(self):
        self.call_cmd([self.hostname, "watchdog="])
        self.assert_called_with([_node(self.hostname, watchdog="")])

    def test_with_multiple_watchdogs(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(
                [self.hostname, "watchdog=watchdog1", "watchdog=watchdog2"]
            )
        self.assertEqual(
            (
                "duplicate option 'watchdog' with different values 'watchdog1' "
                "and 'watchdog2'"
            ),
            cm.exception.message,
        )

    def test_with_device(self):
        device = "device"
        self.call_cmd([self.hostname, f"device={device}"])
        self.assert_called_with([_node(self.hostname, devices=[device])])

    def test_with_empty_device(self):
        self.call_cmd([self.hostname, "device="])
        self.assert_called_with([_node(self.hostname, devices=[""])])

    def test_with_multiple_devices(self):
        device_list = [f"device{i}" for i in range(5)]
        self.call_cmd(
            [self.hostname] + [f"device={device}" for device in device_list]
        )
        self.assert_called_with([_node(self.hostname, devices=device_list)])

    def test_with_all_options(self):
        self.call_cmd(
            [
                self.hostname,
                "device=d1",
                "watchdog=w",
                "device=d2",
                "addr=a1",
                "addr=a3",
                "device=d0",
                "addr=a2",
                "addr=a0",
            ]
        )
        self.assert_called_with(
            [
                _node(
                    self.hostname,
                    addrs=["a1", "a3", "a2", "a0"],
                    watchdog="w",
                    devices=["d1", "d2", "d0"],
                )
            ]
        )

    def test_with_unknown_options(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(
                [
                    self.hostname,
                    "unknown=opt",
                    "watchdog=w",
                    "not_supported=opt",
                ]
            )
        self.assertEqual(
            "Unknown options 'not_supported', 'unknown' for node '{}'".format(
                self.hostname
            ),
            cm.exception.message,
        )

    def assert_modifiers(self, modifiers, cmd_params=None):
        if cmd_params is None:
            cmd_params = modifiers
        self.call_cmd([self.hostname], modifiers)
        self.assert_called_with([_node(self.hostname)], **cmd_params)

    def test_enable(self):
        self.assert_modifiers(dict(enable=True))

    def test_start(self):
        self.assert_modifiers(dict(start=True))

    def test_enable_start(self):
        self.assert_modifiers(dict(enable=True, start=True))

    def test_wait(self):
        self.assert_modifiers(dict(wait="10"))

    def test_start_wait(self):
        self.assert_modifiers(dict(start=True, wait=None))

    def test_start_wait_timeout(self):
        self.assert_modifiers(dict(start=True, wait="10"))

    def test_force(self):
        self.assert_modifiers(
            dict(force=True), dict(force_flags=[report_codes.FORCE])
        )

    def test_skip_offline(self):
        self.assert_modifiers(
            {"skip-offline": True},
            dict(force_flags=[report_codes.SKIP_OFFLINE_NODES]),
        )

    def test_no_watchdog_validation(self):
        self.assert_modifiers(
            {"no-watchdog-validation": True}, dict(no_watchdog_validation=True)
        )

    def test_all_modifiers(self):
        self.assert_modifiers(
            {
                "enable": True,
                "start": True,
                "wait": "15",
                "force": True,
                "skip-offline": True,
                "no-watchdog-validation": True,
            },
            dict(
                enable=True,
                start=True,
                wait="15",
                no_watchdog_validation=True,
                force_flags=[
                    report_codes.FORCE,
                    report_codes.SKIP_OFFLINE_NODES,
                ],
            ),
        )


class AddLink(TestCase):
    # pylint: disable=too-many-public-methods
    def setUp(self):
        self.lib = mock.Mock(spec_set=["cluster"])
        self.cluster = mock.Mock(spec_set=["add_link"])
        self.lib.cluster = self.cluster
        self._default_kwargs = dict(
            force_flags=[],
        )

    def assert_called_with(self, link_list, options, **kwargs):
        default_kwargs = dict(self._default_kwargs)
        default_kwargs.update(kwargs)
        self.cluster.add_link.assert_called_once_with(
            link_list, options, **default_kwargs
        )

    def call_cmd(self, argv, modifiers=None):
        cluster.link_add(self.lib, argv, dict_to_modifiers(modifiers or {}))

    def test_no_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd([])
        self.assertIsNone(cm.exception.message)

    def test_addrs(self):
        self.call_cmd(
            ["node1=addr1", "node2=addr2"],
        )
        self.assert_called_with(
            {"node1": "addr1", "node2": "addr2"},
            {},
        )

    def test_options(self):
        self.call_cmd(
            ["options", "a=b", "c=d"],
        )
        self.assert_called_with(
            {},
            {"a": "b", "c": "d"},
        )

    def test_addrs_and_options(self):
        self.call_cmd(
            ["node1=addr1", "node2=addr2", "options", "a=b", "c=d"],
        )
        self.assert_called_with(
            {"node1": "addr1", "node2": "addr2"},
            {"a": "b", "c": "d"},
        )

    def test_missing_node_name(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(
                ["=addr1", "node2=addr2"],
            )
        self.assertEqual("missing key in '=addr1' option", cm.exception.message)

    def test_missing_node_addr1(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(
                ["node1=addr1", "node2"],
            )
        self.assertEqual(
            "missing value of 'node2' option", cm.exception.message
        )

    def test_missing_node_addr2(self):
        self.call_cmd(
            ["node1=addr1", "node2="],
        )
        self.assert_called_with(
            {"node1": "addr1", "node2": ""},
            {},
        )

    def test_duplicate_node_name(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(
                ["node1=a1", "node1=a2"],
            )
        self.assertEqual(
            "duplicate option 'node1' with different values 'a1' and 'a2'",
            cm.exception.message,
        )

    def test_missing_option_name(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(
                ["node1=addr1", "node2=addr2", "options", "=b", "c=d"],
            )
        self.assertEqual("missing key in '=b' option", cm.exception.message)

    def test_missing_option_value1(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(
                ["node1=addr1", "node2=addr2", "options", "a=b", "c"],
            )
        self.assertEqual("missing value of 'c' option", cm.exception.message)

    def test_missing_option_value2(self):
        self.call_cmd(
            ["node1=addr1", "node2=addr2", "options", "a=b", "c="],
        )
        self.assert_called_with(
            {"node1": "addr1", "node2": "addr2"},
            {"a": "b", "c": ""},
        )

    def test_duplicate_options(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(
                ["node1=addr1", "node2=addr2", "options", "a=b", "a=d"],
            )
        self.assertEqual(
            "duplicate option 'a' with different values 'b' and 'd'",
            cm.exception.message,
        )

    def test_keyword_twice(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(
                ["node1=addr1", "options", "a=b", "options", "c=d"],
            )
        self.assertEqual(
            "'options' cannot be used more than once", cm.exception.message
        )

    def test_force(self):
        self.call_cmd(["node1=addr1", "node2=addr2"], {"force": True})
        self.assert_called_with(
            {"node1": "addr1", "node2": "addr2"},
            {},
            force_flags=[report_codes.FORCE],
        )

    def test_skip_offline(self):
        self.call_cmd(["node1=addr1", "node2=addr2"], {"skip-offline": True})
        self.assert_called_with(
            {"node1": "addr1", "node2": "addr2"},
            {},
            force_flags=[report_codes.SKIP_OFFLINE_NODES],
        )

    def test_request_timeout(self):
        self.call_cmd(["node1=addr1", "node2=addr2"], {"request-timeout": "10"})
        self.assert_called_with(
            {"node1": "addr1", "node2": "addr2"}, {}, force_flags=[]
        )

    def test_corosync_conf(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(
                ["node1=addr1", "node2=addr2"],
                {"corosync_conf": "/tmp/corosync.conf"},
            )
        self.assertEqual(
            (
                "Specified option '--corosync_conf' is not supported "
                "in this command"
            ),
            cm.exception.message,
        )

    def test_all_modifiers(self):
        self.call_cmd(
            ["node1=addr1", "node2=addr2"],
            {
                "force": True,
                "request-timeout": "10",
                "skip-offline": True,
            },
        )
        self.assert_called_with(
            {"node1": "addr1", "node2": "addr2"},
            {},
            force_flags=[report_codes.FORCE, report_codes.SKIP_OFFLINE_NODES],
        )

    def test_unsupported_modifier(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(["node1=addr1", "node2=addr2"], {"start": True})
        self.assertEqual(
            "Specified option '--start' is not supported in this command",
            cm.exception.message,
        )


class RemoveLink(TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["cluster"])
        self.cluster = mock.Mock(spec_set=["remove_links"])
        self.lib.cluster = self.cluster
        self._default_kwargs = dict(
            force_flags=[],
        )

    def assert_called_with(self, link_list, **kwargs):
        default_kwargs = dict(self._default_kwargs)
        default_kwargs.update(kwargs)
        self.cluster.remove_links.assert_called_once_with(
            link_list, **default_kwargs
        )

    def call_cmd(self, argv, modifiers=None):
        cluster.link_remove(self.lib, argv, dict_to_modifiers(modifiers or {}))

    def test_no_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd([])
        self.assertIsNone(cm.exception.message)

    def test_one_link(self):
        self.call_cmd(["1"])
        self.assert_called_with(["1"])

    def test_more_links(self):
        self.call_cmd(["1", "2"])
        self.assert_called_with(["1", "2"])

    def test_skip_offline(self):
        self.call_cmd(["1"], {"skip-offline": True})
        self.assert_called_with(
            ["1"], force_flags=[report_codes.SKIP_OFFLINE_NODES]
        )

    def test_request_timeout(self):
        self.call_cmd(["1"], {"request-timeout": "10"})
        self.assert_called_with(["1"], force_flags=[])

    def test_corosync_conf(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(["1"], {"corosync_conf": "/tmp/corosync.conf"})
        self.assertEqual(
            (
                "Specified option '--corosync_conf' is not supported "
                "in this command"
            ),
            cm.exception.message,
        )

    def test_all_modifiers(self):
        self.call_cmd(
            ["1"],
            {
                "skip-offline": True,
                "request-timeout": "10",
            },
        )
        self.assert_called_with(
            ["1"], force_flags=[report_codes.SKIP_OFFLINE_NODES]
        )

    def test_unsupported_modifier(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(["1"], {"start": True})
        self.assertEqual(
            "Specified option '--start' is not supported in this command",
            cm.exception.message,
        )


class UpdateLink(TestCase):
    # pylint: disable=too-many-public-methods
    def setUp(self):
        self.lib = mock.Mock(spec_set=["cluster"])
        self.cluster = mock.Mock(spec_set=["update_link"])
        self.lib.cluster = self.cluster
        self._default_kwargs = dict(
            force_flags=[],
        )

    def assert_called_with(self, linknumber, addrs, options, **kwargs):
        default_kwargs = dict(self._default_kwargs)
        default_kwargs.update(kwargs)
        self.cluster.update_link.assert_called_once_with(
            linknumber, addrs, options, **default_kwargs
        )

    def call_cmd(self, argv, modifiers=None):
        cluster.link_update(self.lib, argv, dict_to_modifiers(modifiers or {}))

    def test_no_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd([])
        self.assertIsNone(cm.exception.message)

    def test_one_arg(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(["0"])
        self.assertIsNone(cm.exception.message)

    def test_addrs(self):
        self.call_cmd(
            ["0", "node1=addr1", "node2=addr2"],
        )
        self.assert_called_with(
            "0",
            {"node1": "addr1", "node2": "addr2"},
            {},
        )

    def test_options(self):
        self.call_cmd(
            ["0", "options", "a=b", "c=d"],
        )
        self.assert_called_with(
            "0",
            {},
            {"a": "b", "c": "d"},
        )

    def test_addrs_and_options(self):
        self.call_cmd(
            ["1", "node1=addr1", "node2=addr2", "options", "a=b", "c=d"],
        )
        self.assert_called_with(
            "1",
            {"node1": "addr1", "node2": "addr2"},
            {"a": "b", "c": "d"},
        )

    def test_missing_node_name(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(
                ["0", "=addr1", "node2=addr2"],
            )
        self.assertEqual("missing key in '=addr1' option", cm.exception.message)

    def test_missing_node_addr1(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(
                ["0", "node1=addr1", "node2"],
            )
        self.assertEqual(
            "missing value of 'node2' option", cm.exception.message
        )

    def test_missing_node_addr2(self):
        self.call_cmd(
            ["0", "node1=addr1", "node2="],
        )
        self.assert_called_with(
            "0",
            {"node1": "addr1", "node2": ""},
            {},
        )

    def test_duplicate_node_name(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(
                ["0", "node1=a1", "node1=a2"],
            )
        self.assertEqual(
            "duplicate option 'node1' with different values 'a1' and 'a2'",
            cm.exception.message,
        )

    def test_missing_option_name(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(
                ["0", "node1=addr1", "node2=addr2", "options", "=b", "c=d"],
            )
        self.assertEqual("missing key in '=b' option", cm.exception.message)

    def test_missing_option_value1(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(
                ["0", "node1=addr1", "node2=addr2", "options", "a=b", "c"],
            )
        self.assertEqual("missing value of 'c' option", cm.exception.message)

    def test_missing_option_value2(self):
        self.call_cmd(
            ["0", "node1=addr1", "node2=addr2", "options", "a=b", "c="],
        )
        self.assert_called_with(
            "0",
            {"node1": "addr1", "node2": "addr2"},
            {"a": "b", "c": ""},
        )

    def test_duplicate_options(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(
                ["0", "node1=addr1", "node2=addr2", "options", "a=b", "a=d"],
            )
        self.assertEqual(
            "duplicate option 'a' with different values 'b' and 'd'",
            cm.exception.message,
        )

    def test_keyword_twice(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(
                ["0", "node1=addr1", "options", "a=b", "options", "c=d"],
            )
        self.assertEqual(
            "'options' cannot be used more than once", cm.exception.message
        )

    def test_force(self):
        self.call_cmd(["1", "node1=addr1", "node2=addr2"], {"force": True})
        self.assert_called_with(
            "1",
            {"node1": "addr1", "node2": "addr2"},
            {},
            force_flags=[report_codes.FORCE],
        )

    def test_skip_offline(self):
        self.call_cmd(
            ["1", "node1=addr1", "node2=addr2"], {"skip-offline": True}
        )
        self.assert_called_with(
            "1",
            {"node1": "addr1", "node2": "addr2"},
            {},
            force_flags=[report_codes.SKIP_OFFLINE_NODES],
        )

    def test_request_timeout(self):
        self.call_cmd(
            ["2", "node1=addr1", "node2=addr2"], {"request-timeout": "10"}
        )
        self.assert_called_with(
            "2", {"node1": "addr1", "node2": "addr2"}, {}, force_flags=[]
        )

    def test_corosync_conf(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(
                ["1", "node1=addr1", "node2=addr2"],
                {"corosync_conf": "/tmp/corosync.conf"},
            )
        self.assertEqual(
            (
                "Specified option '--corosync_conf' is not supported "
                "in this command"
            ),
            cm.exception.message,
        )

    def test_all_modifiers(self):
        self.call_cmd(
            ["0", "node1=addr1", "node2=addr2"],
            {
                "force": True,
                "request-timeout": "10",
                "skip-offline": True,
            },
        )
        self.assert_called_with(
            "0",
            {"node1": "addr1", "node2": "addr2"},
            {},
            force_flags=[report_codes.FORCE, report_codes.SKIP_OFFLINE_NODES],
        )

    def test_unsupported_modifier(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(["0", "node1=addr1", "node2=addr2"], {"start": True})
        self.assertEqual(
            "Specified option '--start' is not supported in this command",
            cm.exception.message,
        )


class ConfigUpdate(TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["cluster"])
        self.lib.cluster = mock.Mock(
            spec_set=["config_update", "config_update_local"],
        )

    def assert_update_called_with(self, transport, compression, crypto, totem):
        self.lib.cluster.config_update.assert_called_once_with(
            transport, compression, crypto, totem
        )
        self.lib.cluster.config_update_local.assert_not_called()

    def assert_config_update_local_called_with(
        self, transport, compression, crypto, totem
    ):
        self.lib.cluster.config_update_local.assert_called_once_with(
            b"", transport, compression, crypto, totem
        )
        self.lib.cluster.config_update.assert_not_called()

    def call_cmd(self, argv, modifiers=None):
        cluster.config_update(
            self.lib, argv, dict_to_modifiers(modifiers or {})
        )

    def test_no_args(self):
        self.call_cmd([])
        self.assert_update_called_with({}, {}, {}, {})

    def test_unknown_keyword(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(["unknown_keyword"])
        self.assertIsNone(cm.exception.message)

    def test_unknown_keyword_with_options(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(
                ["unknown_keyword", "a=b", "c=d", "compression", "e=f", "g=h"],
            )
        self.assertIsNone(cm.exception.message)

    def test_supported_keywords_without_options(self):
        self.call_cmd(["transport", "compression", "crypto", "totem"])
        self.assert_update_called_with({}, {}, {}, {})

    def test_supported_keywords_with_options(self):
        self.call_cmd(
            [
                "transport",
                "a=b",
                "c=d",
                "compression",
                "e=f",
                "g=h",
                "crypto",
                "i=j",
                "k=l",
                "totem",
                "m=n",
                "o=p",
            ],
        )
        self.assert_update_called_with(
            {"a": "b", "c": "d"},
            {"e": "f", "g": "h"},
            {"i": "j", "k": "l"},
            {"m": "n", "o": "p"},
        )

    def test_repeated_keywords_with_options(self):
        self.call_cmd(
            [
                "crypto",
                "a=b",
                "c=d",
                "totem",
                "i=j",
                "k=l",
                "crypto",
                "e=f",
                "g=h",
                "totem",
                "m=n",
                "o=p",
            ],
        )
        self.assert_update_called_with(
            {},
            {},
            {"a": "b", "c": "d", "e": "f", "g": "h"},
            {"i": "j", "k": "l", "m": "n", "o": "p"},
        )

    def test_missing_option_name(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(
                ["transport", "=b", "c=d", "compression", "e=f", "=h"],
            )
        self.assertEqual("missing key in '=b' option", cm.exception.message)

    def test_missing_option_value(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(
                ["transport", "a=b", "c=d", "compression", "e", "g=h"],
            )
        self.assertEqual("missing value of 'e' option", cm.exception.message)

    def test_empty_option_values_allowed(self):
        self.call_cmd(
            ["transport", "a=", "c=d", "crypto", "e=f", "g="],
        )
        self.assert_update_called_with(
            {"a": "", "c": "d"}, {}, {"e": "f", "g": ""}, {}
        )

    def test_duplicate_options(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(["transport", "a=b", "a=c", "crypto", "b=c", "b=d"])
        self.assertEqual(
            "duplicate option 'a' with different values 'b' and 'c'",
            cm.exception.message,
        )

    def test_duplicate_options_same_value(self):
        self.call_cmd(["transport", "a=b", "a=b", "crypto", "b=c", "b=c"])
        self.assert_update_called_with({"a": "b"}, {}, {"b": "c"}, {})

    def test_unsupported_modifier(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(["transport", "a=b", "c=d"], {"skip-offline": True})
        self.assertEqual(
            (
                "Specified option '--skip-offline' is not supported in this "
                "command"
            ),
            cm.exception.message,
        )

    def test_corosync_conf_modifier(self):
        corosync_conf_data = b"new corosync.conf"
        self.lib.cluster.config_update_local.return_value = corosync_conf_data
        with get_tmp_file("test_cluster_corosync.conf", "rb") as output_file:
            self.call_cmd(
                [
                    "transport",
                    "a=b",
                    "c=d",
                    "compression",
                    "e=f",
                    "g=h",
                    "crypto",
                    "i=j",
                    "k=l",
                    "totem",
                    "m=n",
                    "o=p",
                ],
                {"corosync_conf": output_file.name},
            )
            self.assertEqual(output_file.read(), corosync_conf_data)
        self.assert_config_update_local_called_with(
            {"a": "b", "c": "d"},
            {"e": "f", "g": "h"},
            {"i": "j", "k": "l"},
            {"m": "n", "o": "p"},
        )


@mock.patch("pcs.cluster.print")
class ConfigShow(TestCase):
    @staticmethod
    def fixture_corosync_dto(with_qdevice=True):
        qdevice = None
        if with_qdevice:
            qdevice = CorosyncQuorumDeviceSettingsDto(
                model="net",
                model_options={"algorithm": "ffsplit", "host": "node-qdevice"},
                generic_options={"sync_timeout": "5000", "timeout": "5000"},
                heuristics_options={
                    "mode": "on",
                    "exec_ping": "/usr/bin/ping -c 1 127.0.0.1",
                },
            )

        return CorosyncConfDto(
            cluster_name="HACluster",
            transport=CorosyncTransportType.KNET,
            totem_options={"census": "3600", "join": "50", "token": "3000"},
            transport_options={"ip_version": "ipv4-6", "link_mode": "passive"},
            compression_options={
                "level": "5",
                "model": "zlib",
                "threshold": "100",
            },
            crypto_options={"cipher": "aes256", "hash": "sha256"},
            nodes=[
                CorosyncNodeDto(
                    name="node2",
                    nodeid="2",
                    addrs=[
                        CorosyncNodeAddressDto(
                            addr="10.0.0.2",
                            link="2",
                            type="IPv4",
                        ),
                        CorosyncNodeAddressDto(
                            addr="node2",
                            link="0",
                            type="FQDN",
                        ),
                    ],
                ),
                CorosyncNodeDto(
                    name="node1",
                    nodeid="1",
                    addrs=[
                        CorosyncNodeAddressDto(
                            addr="node1",
                            link="0",
                            type="FQDN",
                        ),
                        CorosyncNodeAddressDto(
                            addr="10.0.0.1",
                            link="2",
                            type="IPv4",
                        ),
                    ],
                ),
            ],
            links_options={
                "0": {
                    "linknumber": "0",
                    "link_priority": "100",
                    "ping_interval": "750",
                    "ping_timeout": "1500",
                    "transport": "udp",
                },
                "1": {
                    "linknumber": "1",
                    "link_priority": "200",
                    "ping_interval": "750",
                    "ping_timeout": "1500",
                    "transport": "udp",
                },
            },
            quorum_options={
                "last_man_standing": "1",
                "last_man_standing_window": "1000",
            },
            quorum_device=qdevice,
        )

    def setUp(self):
        self.lib_call = mock.Mock()
        self.lib = mock.Mock(spec_set=["cluster"])
        self.lib.cluster = mock.Mock(spec_set=["get_corosync_conf_struct"])
        self.lib.cluster.get_corosync_conf_struct = self.lib_call

        self.lib_call.return_value = self.fixture_corosync_dto()
        self.output_text = dedent(
            """\
            Cluster Name: HACluster
            Transport: knet
            Nodes:
              node1:
                Link 0 address: node1
                Link 2 address: 10.0.0.1
                nodeid: 1
              node2:
                Link 0 address: node2
                Link 2 address: 10.0.0.2
                nodeid: 2
            Links:
              Link 0:
                link_priority: 100
                linknumber: 0
                ping_interval: 750
                ping_timeout: 1500
                transport: udp
              Link 1:
                link_priority: 200
                linknumber: 1
                ping_interval: 750
                ping_timeout: 1500
                transport: udp
            Transport Options:
              ip_version: ipv4-6
              link_mode: passive
            Compression Options:
              level: 5
              model: zlib
              threshold: 100
            Crypto Options:
              cipher: aes256
              hash: sha256
            Totem Options:
              census: 3600
              join: 50
              token: 3000
            Quorum Options:
              last_man_standing: 1
              last_man_standing_window: 1000
            Quorum Device: net
              Options:
                sync_timeout: 5000
                timeout: 5000
              Model Options:
                algorithm: ffsplit
                host: node-qdevice
              Heuristics:
                exec_ping: /usr/bin/ping -c 1 127.0.0.1
                mode: on"""
        )

    def call_cmd(self, argv, modifiers=None):
        cluster.config_show(self.lib, argv, dict_to_modifiers(modifiers or {}))

    def test_args_not_allowd(self, mock_print):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(["arg"])
        self.assertIsNone(cm.exception.message)
        self.lib_call.assert_not_called()
        mock_print.assert_not_called()

    def test_unsupported_option(self, mock_print):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd([], {"unsupported-option": True})
        self.assertEqual(
            (
                "Specified option '--unsupported-option' is not supported in "
                "this command"
            ),
            cm.exception.message,
        )
        self.lib_call.assert_not_called()
        mock_print.assert_not_called()

    def test_output_format_unknown(self, mock_print):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd([], {"output-format": "unknown"})
        self.assertEqual(
            (
                "Unknown value 'unknown' for '--output-format' option. "
                "Supported values are: 'cmd', 'json', 'text'"
            ),
            cm.exception.message,
        )
        self.lib_call.assert_not_called()
        mock_print.assert_not_called()

    def test_output_format_default(self, mock_print):
        self.call_cmd([], {"corosync_conf": "some_file_name"})
        self.lib_call.assert_called_once_with()
        mock_print.assert_called_once_with(self.output_text)

    def test_output_format_text(self, mock_print):
        self.call_cmd([], {"output-format": "text"})
        self.lib_call.assert_called_once_with()
        mock_print.assert_called_once_with(self.output_text)

    def test_output_format_json(self, mock_print):
        self.call_cmd([], {"output-format": "json"})
        self.lib_call.assert_called_once_with()
        mock_print.assert_called_once_with(
            json.dumps(dto.to_dict(self.lib_call.return_value))
        )

    cmd_output = dedent(
        """\
        pcs cluster setup HACluster \\
          node1 addr=node1 addr=10.0.0.1 \\
          node2 addr=node2 addr=10.0.0.2 \\
          transport \\
          knet \\
              ip_version=ipv4-6 \\
              link_mode=passive \\
            link \\
              link_priority=100 \\
              linknumber=0 \\
              ping_interval=750 \\
              ping_timeout=1500 \\
              transport=udp \\
            link \\
              link_priority=200 \\
              linknumber=1 \\
              ping_interval=750 \\
              ping_timeout=1500 \\
              transport=udp \\
            compression \\
              level=5 \\
              model=zlib \\
              threshold=100 \\
            crypto \\
              cipher=aes256 \\
              hash=sha256 \\
          totem \\
            census=3600 \\
            join=50 \\
            token=3000 \\
          quorum \\
            last_man_standing=1 \\
            last_man_standing_window=1000"""
    )

    @mock.patch("pcs.cluster.warn")
    def test_output_format_cmd_with_qdevice(self, mock_warn, mock_print):
        self.call_cmd([], {"output-format": "cmd"})
        self.lib_call.assert_called_once_with()
        mock_print.assert_called_once_with(self.cmd_output)
        mock_warn.assert_called_once_with(
            "Quorum device configuration detected but not yet supported by "
            "this command."
        )

    @mock.patch("pcs.cluster.warn")
    def test_output_format_cmd(self, mock_warn, mock_print):
        self.lib_call.return_value = self.fixture_corosync_dto(
            with_qdevice=False
        )
        self.call_cmd([], {"output-format": "cmd"})
        self.lib_call.assert_called_once_with()
        mock_print.assert_called_once_with(self.cmd_output)
        mock_warn.assert_not_called()


class ClusterAuthkeyCorosync(TestCase):
    def setUp(self):
        self.lib_call = mock.Mock()
        self.lib = mock.Mock(spec_set=["cluster"])
        self.lib.cluster = mock.Mock(spec_set=["corosync_authkey_change"])
        self.lib.cluster.corosync_authkey_change = self.lib_call

    def call_cmd(self, argv, modifiers=None):
        cluster.authkey_corosync(
            self.lib, argv, dict_to_modifiers(modifiers or {})
        )

    def test_noargs(self):
        self.call_cmd([])
        self.lib_call.assert_called_once_with(
            corosync_authkey=None, force_flags=[]
        )

    @mock.patch(
        "pcs.cluster.open", new_callable=mock.mock_open, read_data=b"authkey"
    )
    def test_key_path(self, mock_open):
        self.call_cmd(["/tmp/authkey"])
        self.lib_call.assert_called_once_with(
            corosync_authkey=b"authkey", force_flags=[]
        )
        mock_open.assert_called_once_with("/tmp/authkey", "rb")

    @mock.patch(
        "pcs.cluster.open", new_callable=mock.mock_open, read_data=b"authkey"
    )
    @mock.patch("pcs.utils.err")
    def test_key_path_nonexistent_file(self, mock_err, mock_open):
        mock_open.side_effect = EnvironmentError()
        self.call_cmd(["/tmp/nonexistent"])
        mock_err.assert_called_once_with(
            "Unable to read file '/tmp/nonexistent': None"
        )
        mock_open.assert_called_once_with("/tmp/nonexistent", "rb")

    def test_more_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(["arg1", "arg2"])
        self.assertIsNone(cm.exception.message)
        self.lib_call.assert_not_called()

    def test_force(self):
        self.call_cmd([], {"force": True})
        self.lib_call.assert_called_once_with(
            corosync_authkey=None, force_flags=[report_codes.FORCE]
        )

    @mock.patch(
        "pcs.cluster.open", new_callable=mock.mock_open, read_data=b"authkey"
    )
    def test_key_path_force(self, mock_open):
        self.call_cmd(["path"], {"force": True})
        self.lib_call.assert_called_once_with(
            corosync_authkey=b"authkey", force_flags=[report_codes.FORCE]
        )
        mock_open.assert_called_once_with("path", "rb")

    @mock.patch(
        "pcs.cluster.open", new_callable=mock.mock_open, read_data=b"authkey"
    )
    def test_all_options(self, mock_open):
        self.call_cmd(
            ["path"],
            {"force": True, "skip-offline": True, "request-timeout": "10"},
        )
        self.lib_call.assert_called_once_with(
            corosync_authkey=b"authkey",
            force_flags=[report_codes.FORCE, report_codes.SKIP_OFFLINE_NODES],
        )
        mock_open.assert_called_once_with("path", "rb")

    def test_unsupported_option(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd([], {"wait": 10})
        self.assertEqual(
            "Specified option '--wait' is not supported in this command",
            cm.exception.message,
        )
        self.lib_call.assert_not_called()



from unittest import TestCase

from lxml import etree

from pcs.lib.pacemaker import simulate

from pcs_test.tools.misc import get_test_resource as rc


class GetOperationsFromTransitions(TestCase):
    def test_transitions1(self):
        transitions = etree.parse(rc("transitions01.xml"))
        self.assertEqual(
            [
                simulate.SimulationOperation(
                    operation_id=17,
                    primitive_id="dummy",
                    primitive_long_id="dummy",
                    operation_type=simulate.OPERATION_STOP,
                    on_node="rh7-3",
                ),
                simulate.SimulationOperation(
                    operation_id=18,
                    primitive_id="dummy",
                    primitive_long_id="dummy",
                    operation_type=simulate.OPERATION_START,
                    on_node="rh7-2",
                ),
                simulate.SimulationOperation(
                    operation_id=22,
                    primitive_id="d0",
                    primitive_long_id="d0:1",
                    operation_type=simulate.OPERATION_STOP,
                    on_node="rh7-1",
                ),
                simulate.SimulationOperation(
                    operation_id=23,
                    primitive_id="d0",
                    primitive_long_id="d0:1",
                    operation_type=simulate.OPERATION_START,
                    on_node="rh7-2",
                ),
                simulate.SimulationOperation(
                    operation_id=29,
                    primitive_id="state",
                    primitive_long_id="state:0",
                    operation_type=simulate.OPERATION_STOP,
                    on_node="rh7-3",
                ),
                simulate.SimulationOperation(
                    operation_id=30,
                    primitive_id="state",
                    primitive_long_id="state:0",
                    operation_type=simulate.OPERATION_START,
                    on_node="rh7-2",
                ),
            ],
            simulate.get_operations_from_transitions(transitions),
        )

    def test_transitions2(self):
        transitions = etree.parse(rc("transitions02.xml"))
        self.assertEqual(
            [
                simulate.SimulationOperation(
                    operation_id=26,
                    primitive_id="RemoteNode",
                    primitive_long_id="RemoteNode",
                    operation_type=simulate.OPERATION_STOP,
                    on_node="virt-143",
                ),
                simulate.SimulationOperation(
                    operation_id=29,
                    primitive_id="RemoteNode",
                    primitive_long_id="RemoteNode",
                    operation_type=simulate.OPERATION_MIGRATE_TO,
                    on_node="virt-143",
                ),
                simulate.SimulationOperation(
                    operation_id=30,
                    primitive_id="RemoteNode",
                    primitive_long_id="RemoteNode",
                    operation_type=simulate.OPERATION_MIGRATE_FROM,
                    on_node="virt-142",
                ),
                simulate.SimulationOperation(
                    operation_id=45,
                    primitive_id="dummy8",
                    primitive_long_id="dummy8",
                    operation_type=simulate.OPERATION_STOP,
                    on_node="virt-143",
                ),
                simulate.SimulationOperation(
                    operation_id=46,
                    primitive_id="dummy8",
                    primitive_long_id="dummy8",
                    operation_type=simulate.OPERATION_START,
                    on_node="virt-142",
                ),
            ],
            simulate.get_operations_from_transitions(transitions),
        )


class GetResourcesFromOperations(TestCase):
    operations = [
        simulate.SimulationOperation(
            operation_id=0,
            primitive_id="dummy2",
            primitive_long_id="dummy2:1",
            operation_type=simulate.OPERATION_STOP,
            on_node="node1",
        ),
        simulate.SimulationOperation(
            operation_id=1,
            primitive_id="dummy1",
            primitive_long_id="dummy1",
            operation_type=simulate.OPERATION_STOP,
            on_node="node3",
        ),
        simulate.SimulationOperation(
            operation_id=2,
            primitive_id="dummy1",
            primitive_long_id="dummy1",
            operation_type=simulate.OPERATION_START,
            on_node="node2",
        ),
    ]

    def test_no_operations(self):
        self.assertEqual([], simulate.get_resources_from_operations([]))

    def test_no_operations_exclude(self):
        self.assertEqual(
            [],
            simulate.get_resources_from_operations(
                [], exclude_resources={"dummy1"}
            ),
        )

    def test_some_operations(self):
        self.assertEqual(
            ["dummy1", "dummy2"],
            simulate.get_resources_from_operations(self.operations),
        )

    def test_some_operations_exclude(self):
        self.assertEqual(
            ["dummy2"],
            simulate.get_resources_from_operations(
                self.operations,
                exclude_resources={"dummy1", "dummy2:1", "dummyX"},
            ),
        )


class GetResourcesLeftStoppedDemotedMixin:
    def test_no_operations(self):
        self.assertEqual([], self.call([]))

    def test_down(self):
        self.assertEqual(
            ["dummy"],
            self.call(
                [
                    simulate.SimulationOperation(
                        operation_id=0,
                        primitive_id="dummy",
                        primitive_long_id="dummy",
                        operation_type=self.action_down,
                        on_node="node3",
                    ),
                ]
            ),
        )

    def test_up(self):
        self.assertEqual(
            [],
            self.call(
                [
                    simulate.SimulationOperation(
                        operation_id=0,
                        primitive_id="dummy",
                        primitive_long_id="dummy",
                        operation_type=self.action_up,
                        on_node="node3",
                    ),
                ]
            ),
        )

    def test_down_up(self):
        self.assertEqual(
            [],
            self.call(
                [
                    simulate.SimulationOperation(
                        operation_id=0,
                        primitive_id="dummy",
                        primitive_long_id="dummy",
                        operation_type=self.action_down,
                        on_node="node2",
                    ),
                    simulate.SimulationOperation(
                        operation_id=1,
                        primitive_id="dummy",
                        primitive_long_id="dummy",
                        operation_type=self.action_up,
                        on_node="node3",
                    ),
                ]
            ),
        )

    def test_up_down(self):
        self.assertEqual(
            [],
            self.call(
                [
                    simulate.SimulationOperation(
                        operation_id=0,
                        primitive_id="dummy",
                        primitive_long_id="dummy",
                        operation_type=self.action_up,
                        on_node="node2",
                    ),
                    simulate.SimulationOperation(
                        operation_id=1,
                        primitive_id="dummy",
                        primitive_long_id="dummy",
                        operation_type=self.action_down,
                        on_node="node3",
                    ),
                ]
            ),
        )

    def test_mixed(self):
        self.assertEqual(
            ["dummy1", "dummy2"],
            self.call(
                [
                    simulate.SimulationOperation(
                        operation_id=0,
                        primitive_id="dummy1",
                        primitive_long_id="dummy1",
                        operation_type=self.action_down,
                        on_node="node3",
                    ),
                    simulate.SimulationOperation(
                        operation_id=1,
                        primitive_id="dummy2",
                        primitive_long_id="dummy2",
                        operation_type=self.action_down,
                        on_node="node3",
                    ),
                    simulate.SimulationOperation(
                        operation_id=2,
                        primitive_id="dummy3",
                        primitive_long_id="dummy3",
                        operation_type=self.action_down,
                        on_node="node3",
                    ),
                    simulate.SimulationOperation(
                        operation_id=3,
                        primitive_id="dummy3",
                        primitive_long_id="dummy3",
                        operation_type=self.action_up,
                        on_node="node2",
                    ),
                ]
            ),
        )

    def test_exclude(self):
        self.assertEqual(
            ["dummy2"],
            self.call(
                [
                    simulate.SimulationOperation(
                        operation_id=0,
                        primitive_id="dummy1",
                        primitive_long_id="dummy1",
                        operation_type=self.action_down,
                        on_node="node3",
                    ),
                    simulate.SimulationOperation(
                        operation_id=1,
                        primitive_id="dummy2",
                        primitive_long_id="dummy2",
                        operation_type=self.action_down,
                        on_node="node3",
                    ),
                ],
                exclude_resources={"dummy1", "dummyX"},
            ),
        )


class GetResourcesLeftStopped(GetResourcesLeftStoppedDemotedMixin, TestCase):
    action_up = simulate.OPERATION_START
    action_down = simulate.OPERATION_STOP
    call = staticmethod(simulate.get_resources_left_stopped)

    def test_clone_move(self):
        self.assertEqual(
            [],
            self.call(
                [
                    simulate.SimulationOperation(
                        operation_id=0,
                        primitive_id="dummy1",
                        primitive_long_id="dummy1:0",
                        operation_type=self.action_down,
                        on_node="node3",
                    ),
                    simulate.SimulationOperation(
                        operation_id=1,
                        primitive_id="dummy1",
                        primitive_long_id="dummy1:1",
                        operation_type=self.action_down,
                        on_node="node1",
                    ),
                    simulate.SimulationOperation(
                        operation_id=2,
                        primitive_id="dummy1",
                        primitive_long_id="dummy1:0",
                        operation_type=self.action_up,
                        on_node="node2",
                    ),
                    simulate.SimulationOperation(
                        operation_id=3,
                        primitive_id="dummy1",
                        primitive_long_id="dummy1:1",
                        operation_type=self.action_up,
                        on_node="node4",
                    ),
                ]
            ),
        )

    def test_clone_stop(self):
        self.assertEqual(
            ["dummy1"],
            self.call(
                [
                    simulate.SimulationOperation(
                        operation_id=0,
                        primitive_id="dummy1",
                        primitive_long_id="dummy1:0",
                        operation_type=self.action_down,
                        on_node="node3",
                    ),
                    simulate.SimulationOperation(
                        operation_id=1,
                        primitive_id="dummy1",
                        primitive_long_id="dummy1:1",
                        operation_type=self.action_down,
                        on_node="node1",
                    ),
                    simulate.SimulationOperation(
                        operation_id=3,
                        primitive_id="dummy1",
                        primitive_long_id="dummy1:1",
                        operation_type=self.action_up,
                        on_node="node4",
                    ),
                ]
            ),
        )


class GetResourcesLeftDemoted(GetResourcesLeftStoppedDemotedMixin, TestCase):
    action_up = simulate.OPERATION_PROMOTE
    action_down = simulate.OPERATION_DEMOTE
    call = staticmethod(simulate.get_resources_left_demoted)

    def test_master_move(self):
        self.assertEqual(
            [],
            self.call(
                [
                    simulate.SimulationOperation(
                        operation_id=0,
                        primitive_id="dummy1",
                        primitive_long_id="dummy1:0",
                        operation_type=self.action_down,
                        on_node="node3",
                    ),
                    simulate.SimulationOperation(
                        operation_id=3,
                        primitive_id="dummy1",
                        primitive_long_id="dummy1:1",
                        operation_type=self.action_up,
                        on_node="node4",
                    ),
                ]
            ),
        )

    def test_master_stop(self):
        self.assertEqual(
            ["dummy1"],
            self.call(
                [
                    simulate.SimulationOperation(
                        operation_id=0,
                        primitive_id="dummy1",
                        primitive_long_id="dummy1:0",
                        operation_type=self.action_down,
                        on_node="node3",
                    ),
                    simulate.SimulationOperation(
                        operation_id=3,
                        primitive_id="dummy1",
                        primitive_long_id="dummy1:1",
                        operation_type=self.action_up,
                        on_node="node4",
                    ),
                    simulate.SimulationOperation(
                        operation_id=1,
                        primitive_id="dummy1",
                        primitive_long_id="dummy1:2",
                        operation_type=self.action_down,
                        on_node="node1",
                    ),
                ]
            ),
        )



import base64
import binascii
from typing import List

from pcs.common.services.interfaces import ServiceManagerInterface
from pcs.common.services.errors import ManageServiceError

from pcs.common import reports
from pcs.common.reports import (
    codes as report_codes,
    ReportProcessor,
)
from pcs.common.reports.item import (
    get_severity,
    ReportItem,
)
from pcs.lib import external
from pcs.lib.corosync import qdevice_net
from pcs.lib.env import LibraryEnvironment
from pcs.lib.errors import LibraryError
from pcs.lib.services import service_exception_to_report


def qdevice_setup(lib_env: LibraryEnvironment, model, enable, start):
    """
    Initialize qdevice on local host with specified model
    string model qdevice model to initialize
    bool enable make qdevice service start on boot
    bool start start qdevice now
    """
    _check_model(model)
    qdevice_net.qdevice_setup(lib_env.cmd_runner())
    report_processor = lib_env.report_processor
    service_manager = lib_env.service_manager
    report_processor.report(
        ReportItem.info(reports.messages.QdeviceInitializationSuccess(model))
    )
    if enable:
        _service_enable(
            report_processor, service_manager, qdevice_net.SERVICE_NAME
        )
    if start:
        _service_start(
            report_processor, service_manager, qdevice_net.SERVICE_NAME
        )


def qdevice_destroy(lib_env: LibraryEnvironment, model, proceed_if_used=False):
    """
    Stop and disable qdevice on local host and remove its configuration
    string model qdevice model to destroy
    bool procced_if_used destroy qdevice even if it is used by clusters
    """
    _check_model(model)
    report_processor = lib_env.report_processor
    service_manager = lib_env.service_manager
    _check_qdevice_not_used(
        report_processor, lib_env.cmd_runner(), model, proceed_if_used
    )
    _service_stop(report_processor, service_manager, qdevice_net.SERVICE_NAME)
    _service_disable(
        report_processor, service_manager, qdevice_net.SERVICE_NAME
    )
    qdevice_net.qdevice_destroy()
    report_processor.report(
        ReportItem.info(reports.messages.QdeviceDestroySuccess(model))
    )


def qdevice_status_text(
    lib_env: LibraryEnvironment,
    model,
    verbose=False,
    cluster=None,
):
    """
    Get runtime status of a quorum device in plain text
    string model qdevice model to query
    bool verbose get more detailed output
    string cluster show information only about specified cluster
    """
    _check_model(model)
    runner = lib_env.cmd_runner()
    try:
        return qdevice_net.qdevice_status_generic_text(
            runner, verbose
        ) + qdevice_net.qdevice_status_cluster_text(runner, cluster, verbose)
    except qdevice_net.QnetdNotRunningException as e:
        raise LibraryError(
            ReportItem.error(reports.messages.QdeviceNotRunning(model))
        ) from e


def qdevice_enable(lib_env: LibraryEnvironment, model):
    """
    make qdevice start automatically on boot on local host
    """
    _check_model(model)
    _service_enable(
        lib_env.report_processor,
        lib_env.service_manager,
        qdevice_net.SERVICE_NAME,
    )


def qdevice_disable(lib_env: LibraryEnvironment, model):
    """
    make qdevice not start automatically on boot on local host
    """
    _check_model(model)
    _service_disable(
        lib_env.report_processor,
        lib_env.service_manager,
        qdevice_net.SERVICE_NAME,
    )


def qdevice_start(lib_env: LibraryEnvironment, model):
    """
    start qdevice now on local host
    """
    _check_model(model)
    if not qdevice_net.qdevice_initialized():
        raise LibraryError(
            ReportItem.error(reports.messages.QdeviceNotInitialized(model))
        )
    _service_start(
        lib_env.report_processor,
        lib_env.service_manager,
        qdevice_net.SERVICE_NAME,
    )


def qdevice_stop(lib_env: LibraryEnvironment, model, proceed_if_used=False):
    """
    stop qdevice now on local host
    string model qdevice model to destroy
    bool procced_if_used stop qdevice even if it is used by clusters
    """
    _check_model(model)
    _check_qdevice_not_used(
        lib_env.report_processor, lib_env.cmd_runner(), model, proceed_if_used
    )
    _service_stop(
        lib_env.report_processor,
        lib_env.service_manager,
        qdevice_net.SERVICE_NAME,
    )


def qdevice_kill(lib_env: LibraryEnvironment, model):
    """
    kill qdevice now on local host
    """
    _check_model(model)
    _service_kill(lib_env, qdevice_net.SERVICE_NAME)


def qdevice_net_sign_certificate_request(
    lib_env: LibraryEnvironment,
    certificate_request,
    cluster_name,
):
    """
    Sign node certificate request by qnetd CA
    string certificate_request base64 encoded certificate request
    string cluster_name name of the cluster to which qdevice is being added
    """
    try:
        certificate_request_data = base64.b64decode(certificate_request)
    except (TypeError, binascii.Error) as e:
        raise LibraryError(
            ReportItem.error(
                reports.messages.InvalidOptionValue(
                    "qnetd certificate request",
                    certificate_request,
                    ["base64 encoded certificate"],
                )
            )
        ) from e
    return base64.b64encode(
        qdevice_net.qdevice_sign_certificate_request(
            lib_env.cmd_runner(), certificate_request_data, cluster_name
        )
    )


def client_net_setup(lib_env: LibraryEnvironment, ca_certificate):
    """
    Intialize qdevice net client on local host
    ca_certificate base64 encoded qnetd CA certificate
    """
    try:
        ca_certificate_data = base64.b64decode(ca_certificate)
    except (TypeError, binascii.Error) as e:
        raise LibraryError(
            ReportItem.error(
                reports.messages.InvalidOptionValue(
                    "qnetd CA certificate",
                    ca_certificate,
                    ["base64 encoded certificate"],
                )
            )
        ) from e
    qdevice_net.client_setup(lib_env.cmd_runner(), ca_certificate_data)


def client_net_import_certificate(lib_env: LibraryEnvironment, certificate):
    """
    Import qnetd client certificate to local node certificate storage
    certificate base64 encoded qnetd client certificate
    """
    try:
        certificate_data = base64.b64decode(certificate)
    except (TypeError, binascii.Error) as e:
        raise LibraryError(
            ReportItem.error(
                reports.messages.InvalidOptionValue(
                    "qnetd client certificate",
                    certificate,
                    ["base64 encoded certificate"],
                )
            )
        ) from e
    qdevice_net.client_import_certificate_and_key(
        lib_env.cmd_runner(), certificate_data
    )


def client_net_destroy(lib_env: LibraryEnvironment):
    # pylint: disable=unused-argument
    """
    delete qdevice client config files on local host
    """
    qdevice_net.client_destroy()


def _check_model(model):
    if model != "net":
        raise LibraryError(
            ReportItem.error(
                reports.messages.InvalidOptionValue("model", model, ["net"])
            )
        )


def _check_qdevice_not_used(
    reporter: ReportProcessor, runner, model, force=False
):
    _check_model(model)
    connected_clusters: List[str] = []
    if model == "net":
        try:
            status = qdevice_net.qdevice_status_cluster_text(runner)
            connected_clusters = qdevice_net.qdevice_connected_clusters(status)
        except qdevice_net.QnetdNotRunningException:
            pass
    if connected_clusters:
        reporter.report(
            ReportItem(
                severity=get_severity(report_codes.FORCE, force),
                message=reports.messages.QdeviceUsedByClusters(
                    connected_clusters,
                ),
            )
        )
        if reporter.has_errors:
            raise LibraryError()


def _service_start(
    report_processor: ReportProcessor,
    service_manager: ServiceManagerInterface,
    service: str,
) -> None:
    report_processor.report(
        ReportItem.info(
            reports.messages.ServiceActionStarted(
                reports.const.SERVICE_ACTION_START, "quorum device"
            )
        )
    )
    try:
        service_manager.start(service)
    except ManageServiceError as e:
        raise LibraryError(service_exception_to_report(e)) from e
    report_processor.report(
        ReportItem.info(
            reports.messages.ServiceActionSucceeded(
                reports.const.SERVICE_ACTION_START,
                "quorum device",
            )
        )
    )


def _service_stop(
    report_processor: ReportProcessor,
    service_manager: ServiceManagerInterface,
    service: str,
) -> None:
    report_processor.report(
        ReportItem.info(
            reports.messages.ServiceActionStarted(
                reports.const.SERVICE_ACTION_STOP, "quorum device"
            )
        )
    )
    try:
        service_manager.stop(service)
    except ManageServiceError as e:
        raise LibraryError(service_exception_to_report(e)) from e
    report_processor.report(
        ReportItem.info(
            reports.messages.ServiceActionSucceeded(
                reports.const.SERVICE_ACTION_STOP, "quorum device"
            )
        )
    )


def _service_kill(lib_env: LibraryEnvironment, service: str) -> None:
    try:
        external.kill_services(lib_env.cmd_runner(), [service])
    except external.KillServicesError as e:
        raise LibraryError(
            *[
                ReportItem.error(
                    reports.messages.ServiceActionFailed(
                        reports.const.SERVICE_ACTION_KILL, service, e.message
                    )
                )
                for service in e.service
            ]
        ) from e
    lib_env.report_processor.report(
        ReportItem.info(
            reports.messages.ServiceActionSucceeded(
                reports.const.SERVICE_ACTION_KILL, "quorum device"
            )
        )
    )


def _service_enable(
    report_processor: ReportProcessor,
    service_manager: ServiceManagerInterface,
    service: str,
) -> None:
    try:
        service_manager.enable(service)
    except ManageServiceError as e:
        raise LibraryError(service_exception_to_report(e)) from e
    report_processor.report(
        ReportItem.info(
            reports.messages.ServiceActionSucceeded(
                reports.const.SERVICE_ACTION_ENABLE, "quorum device"
            )
        )
    )


def _service_disable(
    report_processor: ReportProcessor,
    service_manager: ServiceManagerInterface,
    service: str,
) -> None:
    try:
        service_manager.disable(service)
    except ManageServiceError as e:
        raise LibraryError(service_exception_to_report(e)) from e
    report_processor.report(
        ReportItem.info(
            reports.messages.ServiceActionSucceeded(
                reports.const.SERVICE_ACTION_DISABLE, "quorum device"
            )
        )
    )



from typing import (
    Any,
    Sequence,
)

from pcs import (
    cluster,
    pcsd,
    resource,
    status,
    usage,
)
import pcs.cli.cluster.command as cluster_command
from pcs.cli.common.errors import (
    CmdLineInputError,
    raise_command_replaced,
)
from pcs.cli.common.parse_args import InputModifiers
from pcs.cli.common.routing import create_router
from pcs.cli.reports.output import warn
from pcs.utils import exit_on_cmdline_input_errror


def certkey(lib: Any, argv: Sequence[str], modifiers: InputModifiers) -> None:
    warn(
        "This command is deprecated and will be removed. "
        "Please use 'pcs pcsd certkey' instead.",
        stderr=True,
    )
    try:
        return pcsd.pcsd_certkey_cmd(lib, argv, modifiers)
    except CmdLineInputError as e:
        return exit_on_cmdline_input_errror(e, "pcsd", ["certkey"])


def pcsd_status(
    lib: Any, argv: Sequence[str], modifiers: InputModifiers
) -> None:
    warn(
        "This command is deprecated and will be removed. "
        "Please use 'pcs pcsd status' or 'pcs status pcsd' instead.",
        stderr=True,
    )
    try:
        return pcsd.pcsd_status_cmd(lib, argv, modifiers)
    except CmdLineInputError as e:
        return exit_on_cmdline_input_errror(e, "pcsd", ["status"])


cluster_cmd = create_router(
    {
        "help": lambda lib, argv, modifiers: usage.cluster(argv),
        "setup": cluster.cluster_setup,
        "config": create_router(
            {
                "show": cluster.config_show,
                "update": cluster.config_update,
            },
            ["cluster", "config"],
            default_cmd="show",
        ),
        "authkey": create_router(
            {"corosync": cluster.authkey_corosync},
            ["cluster", "authkey"],
        ),
        "sync": create_router(
            {
                "corosync": cluster.sync_nodes,
            },
            ["cluster", "sync"],
            default_cmd="corosync",
        ),
        "status": status.cluster_status,
        # TODO remove, deprecated command
        # replaced with 'pcs pcsd status' and 'pcs status pcsd'
        "pcsd-status": pcsd_status,
        # TODO remove, deprecated command
        # replaced with 'pcs pcsd certkey'
        "certkey": certkey,
        "auth": cluster.cluster_auth_cmd,
        "start": cluster.cluster_start_cmd,
        "stop": cluster.cluster_stop_cmd,
        "kill": cluster.kill_cluster,
        "enable": cluster.cluster_enable_cmd,
        "disable": cluster.cluster_disable_cmd,
        "cib": cluster.get_cib,
        "cib-push": cluster.cluster_push,
        "cib-upgrade": cluster.cluster_cib_upgrade_cmd,
        "edit": cluster.cluster_edit,
        "link": create_router(
            {
                "add": cluster.link_add,
                "delete": cluster.link_remove,
                "remove": cluster.link_remove,
                "update": cluster.link_update,
            },
            ["cluster", "link"],
        ),
        "node": create_router(
            {
                "add": cluster.node_add,
                "add-guest": cluster_command.node_add_guest,
                "add-outside": cluster.node_add_outside_cluster,
                "add-remote": cluster_command.node_add_remote,
                "clear": cluster_command.node_clear,
                "delete": cluster.node_remove,
                "delete-guest": cluster_command.node_remove_guest,
                "delete-remote": cluster_command.create_node_remove_remote(
                    resource.resource_remove
                ),
                "remove": cluster.node_remove,
                "remove-guest": cluster_command.node_remove_guest,
                "remove-remote": cluster_command.create_node_remove_remote(
                    resource.resource_remove
                ),
            },
            ["cluster", "node"],
        ),
        "uidgid": cluster.cluster_uidgid,
        "corosync": cluster.cluster_get_corosync_conf,
        "reload": cluster.cluster_reload,
        "destroy": cluster.cluster_destroy,
        "verify": cluster.cluster_verify,
        "report": cluster.cluster_report,
        "remove_nodes_from_cib": cluster.remove_nodes_from_cib,
        # removed commands
        # These print error messages which point users to the changes section in
        # pcs manpage.
        # To be removed in the next significant version.
        "quorum": lambda lib, argv, modifiers: raise_command_replaced(
            "pcs quorum"
        ),
        "remote-node": create_router(
            {
                "add": lambda lib, argv, modifiers: raise_command_replaced(
                    "pcs cluster node add-guest",
                ),
                "remove": lambda lib, argv, modifiers: raise_command_replaced(
                    "pcs cluster node delete-guest",
                    "pcs cluster node remove-guest",
                ),
            },
            ["cluster", "node"],
        ),
        "standby": lambda lib, argv, modifiers: raise_command_replaced(
            "pcs node standby"
        ),
        "unstandby": lambda lib, argv, modifiers: raise_command_replaced(
            "pcs node unstandby"
        ),
    },
    ["cluster"],
)



import socket
import os.path

from functools import lru_cache

from tornado.iostream import IOStream

from pcs import settings
from pcs.daemon import log


async def notify(socket_name):
    if socket_name[0] == "@":
        # abstract namespace socket
        socket_name = "\0" + socket_name[1:]

    log.pcsd.info("Notifying systemd we are running (socket '%s')", socket_name)
    try:
        stream = IOStream(socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM))
        await stream.connect(socket_name)
        await stream.write(b"READY=1")
        stream.close()
        # pylint: disable=broad-except
    except Exception as e:
        log.pcsd.error("Unable to notify systemd on '%s': %s", socket_name, e)


@lru_cache()
def is_systemd():
    return any(os.path.isdir(path) for path in settings.systemd_unit_path)



from unittest import TestCase

from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools
from pcs_test.tools.misc import outdent

from pcs.common.reports import codes as report_codes
from pcs.lib.commands.sbd import get_local_available_watchdogs


def _watchdog_fixture(identity, driver, caution=None):
    info = dict(
        identity=identity,
        driver="<unknown>" if driver is None else driver,
        caution=caution,
    )
    return info


class GetLocalAvailableWatchdogs(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)

    def test_one_device(self):
        self.config.runner.sbd.list_watchdogs(
            outdent(
                """
        Discovered 1 watchdog devices:

        [1] /dev/watchdog
        Identity: i6300ESB timer
        Driver: <unknown>
        """
            )
        )
        self.assertEqual(
            {"/dev/watchdog": _watchdog_fixture("i6300ESB timer", None)},
            get_local_available_watchdogs(self.env_assist.get_env()),
        )

    def test_multiple_devices(self):
        self.config.runner.sbd.list_watchdogs(
            outdent(
                """
        Discovered 3 watchdog devices:

        [1] /dev/watchdog
        Identity: i6300ESB timer
        Driver: <unknown>

        [2] /dev/watchdog2
        Identity: iTCO_wdt
        Driver: iTCO_wdt

        [3] /dev/watchdog3
        Identity: iTCO_wdt
        Driver: iTCO_wdt
        """
            )
        )
        self.assertEqual(
            {
                "/dev/watchdog": _watchdog_fixture("i6300ESB timer", None),
                "/dev/watchdog2": _watchdog_fixture("iTCO_wdt", "iTCO_wdt"),
                "/dev/watchdog3": _watchdog_fixture("iTCO_wdt", "iTCO_wdt"),
            },
            get_local_available_watchdogs(self.env_assist.get_env()),
        )

    def test_with_caution(self):
        self.config.runner.sbd.list_watchdogs(
            outdent(
                """
        Discovered 4 watchdog devices:

        [1] /dev/watchdog
        Identity: i6300ESB timer
        Driver: <unknown>

        [2] /dev/watchdog0
        Identity: Software Watchdog
        Driver: softdog
        CAUTION: Not recommended for use with sbd.

        [3] /dev/watchdog1
        Identity: iTCO_wdt
        Driver: iTCO_wdt

        [4] /dev/watchdog2
        Identity: Software Watchdog
        Driver: softdog
        CAUTION: Not recommended for use with sbd.
        """
            )
        )
        self.assertEqual(
            {
                "/dev/watchdog": _watchdog_fixture("i6300ESB timer", None),
                "/dev/watchdog0": _watchdog_fixture(
                    "Software Watchdog",
                    "softdog",
                    "Not recommended for use with sbd.",
                ),
                "/dev/watchdog1": _watchdog_fixture("iTCO_wdt", "iTCO_wdt"),
                "/dev/watchdog2": _watchdog_fixture(
                    "Software Watchdog",
                    "softdog",
                    "Not recommended for use with sbd.",
                ),
            },
            get_local_available_watchdogs(self.env_assist.get_env()),
        )

    def test_failure(self):
        reason = "some error"
        self.config.runner.sbd.list_watchdogs("", stderr=reason, returncode=1)
        self.env_assist.assert_raise_library_error(
            lambda: get_local_available_watchdogs(self.env_assist.get_env()),
            [
                fixture.error(
                    report_codes.SBD_LIST_WATCHDOG_ERROR,
                    reason=reason,
                )
            ],
            expected_in_processor=False,
        )



import re
from functools import partial

# pylint: disable=protected-access

palete = {
    "black": "\033[30m",
    "red": "\033[31m",
    "green": "\033[32m",
    "orange": "\033[33m",
    "blue": "\033[34m",
    "purple": "\033[35m",
    "cyan": "\033[36m",
    "lightgrey": "\033[37m",
    "darkgrey": "\033[90m",
    "lightred": "\033[91m",
    "lightgreen": "\033[92m",
    "yellow": "\033[93m",
    "lightblue": "\033[94m",
    "pink": "\033[95m",
    "lightcyan": "\033[96m",
    "end": "\033[0m",
    "bold": "\033[1m",
    "underline": "\033[4m",
}

separator1 = "=" * 70
separator2 = "-" * 70

# apply is builtin but is deprecated since 2.3 => no problem to redefine it here
def apply(key_list, text):
    return "".join([palete[key] for key in key_list]) + text + palete["end"]


lightgrey = partial(apply, ["lightgrey"])
bold = partial(apply, ["bold"])
blue = partial(apply, ["blue", "bold"])
red = partial(apply, ["red", "bold"])
green = partial(apply, ["green", "bold"])


def format_module_name(name):
    prefix = ""
    part_list = name.split("_")

    if part_list[0].startswith("test"):
        prefix = "test"
        part_list[0] = part_list[0][len("test") :]

    return prefix + "_".join([bold(part) for part in part_list])


def format_module(test):
    parts = test.__class__.__module__.split(".")
    return lightgrey(".").join(parts[:-1] + [format_module_name(parts[-1])])


def format_test_method_name(test):
    if not hasattr(test, "_testMethodName"):
        return test.id()

    parts = test._testMethodName.split("_")
    if parts[0].startswith("test"):
        parts[0] = lightgrey("test") + parts[0][len("test") :]
    return lightgrey("_").join(parts)


def format_error_overview(errors, failures, slash_last):
    return (
        [
            red(
                "for running failed tests only (errors are first then "
                "failures):"
            ),
            "",
        ]
        + [
            lightgrey(err)
            for err in slash_errors(
                [format_test_name(test) for test, _ in errors + failures],
                slash_last,
            )
        ]
        + [""]
    )


def slash_errors(error_list, slash_last=True):
    if not slash_last:
        return slash_errors(error_list[:-1]) + [error_list[-1]]
    return ["{0} \\".format(err) for err in error_list]


def format_test_name(test):
    if (
        test.__class__.__module__ == "subunit"
        and test.__class__.__name__ == "RemotedTestCase"
    ):
        return format_test_method_name(test)
    return (
        format_module(test)
        + "."
        + test.__class__.__name__
        + "."
        + format_test_method_name(test)
    )


def get_description(test, descriptions):
    doc_first_line = test.shortDescription()
    if descriptions and doc_first_line:
        return "\n".join((str(test), doc_first_line))
    module_parts = test.__class__.__module__.split(".")
    module = module_parts[-1]
    package = ".".join(module_parts[:-1]) + "." if module_parts else ""

    return (
        test._testMethodName
        + " "
        + lightgrey("(")
        + lightgrey(package)
        + bold(module)
        + "."
        + test.__class__.__name__
        + lightgrey(")")
    )


def format_error_list(flavour, errors, descriptions, traceback_highlight):
    line_list = []
    for test, err in errors:
        line_list.extend(
            [
                lightgrey(separator1),
                "%s: %s" % (red(flavour), get_description(test, descriptions)),
                lightgrey(separator2),
                "%s" % format_traceback(err) if traceback_highlight else err,
            ]
        )
    return line_list


def format_traceback(err):
    formated_err = []
    path_regex = re.compile(
        r'^  File "(?P<path>[^"]+)", line (?P<line>\d+), in (?P<name>.*)$'
    )
    was_prev_path = False
    for line in err.splitlines():
        if line == "Traceback (most recent call last):":
            formated_err.append(lightgrey(line))
            was_prev_path = False
            continue

        match = path_regex.match(line)
        if match:
            path = match.group("path").split("/")
            formated_err.append(
                lightgrey('  File "')
                + lightgrey("/").join(path[:-1] + [bold(path[-1])])
                + lightgrey('", line ')
                + bold(match.group("line"))
                + lightgrey(", in ")
                + bold(match.group("name"))
            )
            was_prev_path = True
        elif was_prev_path:
            formated_err.append(bold(line))
            was_prev_path = False
        else:
            formated_err.append(line)
            was_prev_path = False
    return "\n".join(formated_err) + "\n"


def format_skips(skip_map):
    return (
        [blue("Some tests have been skipped:")]
        + [
            lightgrey("{0} ({1}x)".format(reason, len(test_list)))
            for reason, test_list in skip_map.items()
        ]
        + [""]
    )



from textwrap import dedent
from unittest import TestCase

from pcs import settings
from pcs.common import file_type_codes
from pcs.common.reports import codes as report_codes
from pcs.lib.commands import status
from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools
from pcs_test.tools.misc import read_test_resource as rc_read


class FullClusterStatusPlaintext(TestCase):
    # pylint: disable=too-many-public-methods
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.node_name_list = ["node1", "node2", "node3"]
        self.maxDiff = None

    @staticmethod
    def _fixture_xml_clustername(name):
        return """
            <crm_config>
                <cluster_property_set id="cib-bootstrap-options">
                    <nvpair
                        id="cib-bootstrap-options-cluster-name"
                        name="cluster-name" value="{name}"
                    />
                </cluster_property_set>
            </crm_config>
            """.format(
            name=name
        )

    def _fixture_config_live_minimal(self):
        (
            self.config.runner.pcmk.load_state_plaintext(
                stdout="crm_mon cluster status",
            )
            .fs.exists(settings.corosync_conf_file, return_value=True)
            .corosync_conf.load()
            .runner.cib.load(
                resources="""
                <resources>
                    <primitive id="S" class="stonith" type="fence_dummy" />
                </resources>
            """
            )
            .services.is_running(
                "sbd", return_value=False, name="services.is_running.sbd"
            )
        )

    def _fixture_config_live_remote_minimal(self):
        (
            self.config.runner.pcmk.load_state_plaintext(
                stdout="crm_mon cluster status",
            )
            .fs.exists(settings.corosync_conf_file, return_value=False)
            .runner.cib.load(
                optional_in_conf=self._fixture_xml_clustername("test-cib"),
                resources="""
                <resources>
                    <primitive id="S" class="stonith" type="fence_dummy" />
                </resources>
            """,
            )
            .services.is_running(
                "sbd", return_value=False, name="services.is_running.sbd"
            )
        )

    def _fixture_config_local_daemons(
        self,
        corosync_enabled=True,
        corosync_active=True,
        pacemaker_enabled=True,
        pacemaker_active=True,
        pacemaker_remote_enabled=False,
        pacemaker_remote_active=False,
        pcsd_enabled=True,
        pcsd_active=True,
        sbd_enabled=False,
        sbd_active=False,
    ):
        # pylint: disable=too-many-arguments
        (
            self.config.services.is_enabled(
                "corosync",
                name="services.is_enabled.corosync",
                return_value=corosync_enabled,
            )
            .services.is_running(
                "corosync",
                name="services.is_running.corosync",
                return_value=corosync_active,
            )
            .services.is_enabled(
                "pacemaker",
                name="services.is_enabled.pacemaker",
                return_value=pacemaker_enabled,
            )
            .services.is_running(
                "pacemaker",
                name="services.is_running.pacemaker",
                return_value=pacemaker_active,
            )
            .services.is_enabled(
                "pacemaker_remote",
                name="services.is_enabled.pacemaker_remote",
                return_value=pacemaker_remote_enabled,
            )
            .services.is_running(
                "pacemaker_remote",
                name="services.is_running.pacemaker_remote",
                return_value=pacemaker_remote_active,
            )
            .services.is_enabled(
                "pcsd",
                name="services.is_enabled.pcsd",
                return_value=pcsd_enabled,
            )
            .services.is_running(
                "pcsd",
                name="services.is_running.pcsd",
                return_value=pcsd_active,
            )
            .services.is_enabled(
                "sbd",
                name="services.is_enabled.sbd_2",
                return_value=sbd_enabled,
            )
            .services.is_running(
                "sbd",
                name="services.is_running.sbd_2",
                return_value=sbd_active,
            )
        )

    def test_life_cib_mocked_corosync(self):
        self.config.env.set_corosync_conf_data("corosync conf data")
        self.env_assist.assert_raise_library_error(
            lambda: status.full_cluster_status_plaintext(
                self.env_assist.get_env()
            ),
            [
                fixture.error(
                    report_codes.LIVE_ENVIRONMENT_NOT_CONSISTENT,
                    mocked_files=[file_type_codes.COROSYNC_CONF],
                    required_files=[file_type_codes.CIB],
                ),
            ],
            expected_in_processor=False,
        )

    def test_mocked_cib_life_corosync(self):
        self.config.env.set_cib_data("<cib/>")
        self.env_assist.assert_raise_library_error(
            lambda: status.full_cluster_status_plaintext(
                self.env_assist.get_env()
            ),
            [
                fixture.error(
                    report_codes.LIVE_ENVIRONMENT_NOT_CONSISTENT,
                    mocked_files=[file_type_codes.CIB],
                    required_files=[file_type_codes.COROSYNC_CONF],
                ),
            ],
            expected_in_processor=False,
        )

    def test_fail_getting_cluster_status(self):
        (
            self.config.runner.pcmk.load_state_plaintext(
                stdout="some stdout",
                stderr="some stderr",
                returncode=1,
            )
        )
        self.env_assist.assert_raise_library_error(
            lambda: status.full_cluster_status_plaintext(
                self.env_assist.get_env()
            ),
            [
                fixture.error(
                    report_codes.CRM_MON_ERROR,
                    reason="some stderr\nsome stdout",
                ),
            ],
            expected_in_processor=False,
        )

    def test_fail_getting_corosync_conf(self):
        (
            self.config.runner.pcmk.load_state_plaintext(
                stdout="crm_mon cluster status",
            )
            .fs.exists(settings.corosync_conf_file, return_value=True)
            .corosync_conf.load_content("invalid corosync conf")
        )
        self.env_assist.assert_raise_library_error(
            lambda: status.full_cluster_status_plaintext(
                self.env_assist.get_env()
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.PARSE_ERROR_COROSYNC_CONF_LINE_IS_NOT_SECTION_NOR_KEY_VALUE,
                ),
            ]
        )

    def test_fail_getting_cib(self):
        (
            self.config.runner.pcmk.load_state_plaintext(
                stdout="crm_mon cluster status",
            )
            .fs.exists(settings.corosync_conf_file, return_value=True)
            .corosync_conf.load()
            .runner.cib.load_content(
                "some stdout", stderr="cib load error", returncode=1
            )
        )
        self.env_assist.assert_raise_library_error(
            lambda: status.full_cluster_status_plaintext(
                self.env_assist.get_env()
            ),
            [
                fixture.error(
                    report_codes.CIB_LOAD_ERROR,
                    reason="cib load error",
                ),
            ],
            expected_in_processor=False,
        )

    def test_success_live(self):
        self._fixture_config_live_minimal()
        self._fixture_config_local_daemons()
        self.assertEqual(
            status.full_cluster_status_plaintext(self.env_assist.get_env()),
            dedent(
                """\
                Cluster name: test99
                crm_mon cluster status

                Daemon Status:
                  corosync: active/enabled
                  pacemaker: active/enabled
                  pcsd: active/enabled"""
            ),
        )

    def test_success_live_verbose(self):
        (
            self.config.env.set_known_nodes(self.node_name_list)
            .runner.pcmk.can_fence_history_status(stderr="not supported")
            .runner.pcmk.load_state_plaintext(
                verbose=True,
                stdout="crm_mon cluster status",
            )
            .fs.exists(settings.corosync_conf_file, return_value=True)
            .corosync_conf.load(node_name_list=self.node_name_list)
            .runner.cib.load(
                resources="""
                <resources>
                    <primitive id="S" class="stonith" type="fence_dummy" />
                </resources>
            """
            )
            .runner.pcmk.load_ticket_state_plaintext(stdout="ticket status")
            .services.is_running(
                "sbd", return_value=False, name="services.is_running.sbd"
            )
        )
        self._fixture_config_local_daemons()
        (
            self.config.http.host.check_reachability(
                node_labels=self.node_name_list
            )
        )

        self.assertEqual(
            status.full_cluster_status_plaintext(
                self.env_assist.get_env(), verbose=True
            ),
            dedent(
                """\
                Cluster name: test99
                crm_mon cluster status

                Tickets:
                  ticket status

                PCSD Status:
                  node1: Online
                  node2: Online
                  node3: Online

                Daemon Status:
                  corosync: active/enabled
                  pacemaker: active/enabled
                  pcsd: active/enabled"""
            ),
        )

    def test_success_live_remote_node(self):
        self._fixture_config_live_remote_minimal()
        self._fixture_config_local_daemons(
            corosync_enabled=False,
            corosync_active=False,
            pacemaker_enabled=False,
            pacemaker_active=False,
            pacemaker_remote_enabled=True,
            pacemaker_remote_active=True,
        )
        self.assertEqual(
            status.full_cluster_status_plaintext(self.env_assist.get_env()),
            dedent(
                """\
                Cluster name: test-cib
                crm_mon cluster status

                Daemon Status:
                  corosync: inactive/disabled
                  pacemaker: inactive/disabled
                  pacemaker_remote: active/enabled
                  pcsd: active/enabled"""
            ),
        )

    def test_success_live_remote_node_verbose(self):
        (
            self.config.runner.pcmk.can_fence_history_status(
                stderr="not supported"
            )
            .runner.pcmk.load_state_plaintext(
                verbose=True,
                stdout="crm_mon cluster status",
            )
            .fs.exists(settings.corosync_conf_file, return_value=False)
            .runner.cib.load(
                optional_in_conf=self._fixture_xml_clustername("test-cib"),
                resources="""
                <resources>
                    <primitive id="S" class="stonith" type="fence_dummy" />
                </resources>
            """,
            )
            .runner.pcmk.load_ticket_state_plaintext(stdout="ticket status")
            .services.is_running(
                "sbd", return_value=False, name="services.is_running.sbd"
            )
        )
        self._fixture_config_local_daemons(
            corosync_enabled=False,
            corosync_active=False,
            pacemaker_enabled=False,
            pacemaker_active=False,
            pacemaker_remote_enabled=True,
            pacemaker_remote_active=True,
        )

        self.assertEqual(
            status.full_cluster_status_plaintext(
                self.env_assist.get_env(), verbose=True
            ),
            dedent(
                """\
                Cluster name: test-cib
                crm_mon cluster status

                Tickets:
                  ticket status

                Daemon Status:
                  corosync: inactive/disabled
                  pacemaker: inactive/disabled
                  pacemaker_remote: active/enabled
                  pcsd: active/enabled"""
            ),
        )

    def test_succes_mocked(self):
        tmp_file = "/fake/tmp_file"
        env = dict(CIB_file=tmp_file)
        (
            self.config.env.set_corosync_conf_data(rc_read("corosync.conf"))
            .env.set_cib_data("<cib/>", cib_tempfile=tmp_file)
            .runner.pcmk.load_state_plaintext(
                stdout="crm_mon cluster status",
                env=env,
            )
            .runner.cib.load(
                resources="""
                <resources>
                    <primitive id="S" class="stonith" type="fence_dummy" />
                </resources>
            """,
                env=env,
            )
        )
        self.assertEqual(
            status.full_cluster_status_plaintext(self.env_assist.get_env()),
            dedent(
                """\
                Cluster name: test99
                crm_mon cluster status"""
            ),
        )

    def test_succes_mocked_verbose(self):
        tmp_file = "/fake/tmp_file"
        env = dict(CIB_file=tmp_file)
        (
            self.config.env.set_corosync_conf_data(rc_read("corosync.conf"))
            .env.set_cib_data("<cib/>", cib_tempfile=tmp_file)
            .runner.pcmk.can_fence_history_status(
                stderr="not supported",
                env=env,
            )
            .runner.pcmk.load_state_plaintext(
                verbose=True,
                stdout="crm_mon cluster status",
                env=env,
            )
            .runner.cib.load(
                resources="""
                <resources>
                    <primitive id="S" class="stonith" type="fence_dummy" />
                </resources>
            """,
                env=env,
            )
            .runner.pcmk.load_ticket_state_plaintext(
                stdout="ticket status", env=env
            )
        )
        self.assertEqual(
            status.full_cluster_status_plaintext(
                self.env_assist.get_env(), verbose=True
            ),
            dedent(
                """\
                Cluster name: test99
                crm_mon cluster status

                Tickets:
                  ticket status"""
            ),
        )

    def test_success_verbose_inactive_and_fence_history(self):
        (
            self.config.env.set_known_nodes(self.node_name_list)
            .runner.pcmk.can_fence_history_status()
            .runner.pcmk.load_state_plaintext(
                verbose=True,
                inactive=False,
                fence_history=True,
                stdout="crm_mon cluster status",
            )
            .fs.exists(settings.corosync_conf_file, return_value=True)
            .corosync_conf.load(node_name_list=self.node_name_list)
            .runner.cib.load(
                resources="""
                <resources>
                    <primitive id="S" class="stonith" type="fence_dummy" />
                </resources>
            """
            )
            .runner.pcmk.load_ticket_state_plaintext(stdout="ticket status")
            .services.is_running(
                "sbd", return_value=False, name="services.is_running.sbd"
            )
        )
        self._fixture_config_local_daemons()
        (
            self.config.http.host.check_reachability(
                node_labels=self.node_name_list
            )
        )

        self.assertEqual(
            status.full_cluster_status_plaintext(
                self.env_assist.get_env(),
                verbose=True,
                hide_inactive_resources=True,
            ),
            dedent(
                """\
                Cluster name: test99
                crm_mon cluster status

                Tickets:
                  ticket status

                PCSD Status:
                  node1: Online
                  node2: Online
                  node3: Online

                Daemon Status:
                  corosync: active/enabled
                  pacemaker: active/enabled
                  pcsd: active/enabled"""
            ),
        )

    def _assert_success_with_ticket_status_failure(self, stderr="", msg=""):
        (
            self.config.env.set_known_nodes(self.node_name_list)
            .runner.pcmk.can_fence_history_status(stderr="not supported")
            .runner.pcmk.load_state_plaintext(
                verbose=True,
                stdout="crm_mon cluster status",
            )
            .fs.exists(settings.corosync_conf_file, return_value=True)
            .corosync_conf.load(node_name_list=self.node_name_list)
            .runner.cib.load(
                resources="""
                <resources>
                    <primitive id="S" class="stonith" type="fence_dummy" />
                </resources>
            """
            )
            .runner.pcmk.load_ticket_state_plaintext(
                stdout="ticket stdout", stderr=stderr, returncode=1
            )
            .services.is_running(
                "sbd", return_value=False, name="services.is_running.sbd"
            )
        )
        self._fixture_config_local_daemons()
        (
            self.config.http.host.check_reachability(
                node_labels=self.node_name_list
            )
        )

        self.assertEqual(
            status.full_cluster_status_plaintext(
                self.env_assist.get_env(), verbose=True
            ),
            dedent(
                """\
                Cluster name: test99
                crm_mon cluster status

                Tickets:
                  WARNING: Unable to get information about tickets{msg}

                PCSD Status:
                  node1: Online
                  node2: Online
                  node3: Online

                Daemon Status:
                  corosync: active/enabled
                  pacemaker: active/enabled
                  pcsd: active/enabled"""
            ).format(msg=msg),
        )

    def test_success_with_ticket_status_failure(self):
        self._assert_success_with_ticket_status_failure()

    def test_success_with_ticket_status_failure_with_message(self):
        self._assert_success_with_ticket_status_failure(
            stderr="ticket status error\nmultiline\n",
            msg="\n    ticket status error\n    multiline",
        )

    def test_stonith_warning_no_devices(self):
        (
            self.config.runner.pcmk.load_state_plaintext(
                stdout="crm_mon cluster status",
            )
            .fs.exists(settings.corosync_conf_file, return_value=True)
            .corosync_conf.load()
            .runner.cib.load()
            .services.is_running(
                "sbd", return_value=False, name="services.is_running.sbd"
            )
        )
        self._fixture_config_local_daemons()

        self.assertEqual(
            status.full_cluster_status_plaintext(self.env_assist.get_env()),
            dedent(
                """\
                Cluster name: test99

                WARNINGS:
                No stonith devices and stonith-enabled is not false

                crm_mon cluster status

                Daemon Status:
                  corosync: active/enabled
                  pacemaker: active/enabled
                  pcsd: active/enabled"""
            ),
        )

    def test_stonith_warning_no_devices_sbd_enabled(self):
        (
            self.config.runner.pcmk.load_state_plaintext(
                stdout="crm_mon cluster status",
            )
            .fs.exists(settings.corosync_conf_file, return_value=True)
            .corosync_conf.load()
            .runner.cib.load()
            .services.is_running(
                "sbd", return_value=True, name="services.is_running.sbd"
            )
        )
        self._fixture_config_local_daemons()

        self.assertEqual(
            status.full_cluster_status_plaintext(self.env_assist.get_env()),
            dedent(
                """\
                Cluster name: test99
                crm_mon cluster status

                Daemon Status:
                  corosync: active/enabled
                  pacemaker: active/enabled
                  pcsd: active/enabled"""
            ),
        )

    def test_stonith_warnings_regarding_devices_configuration(self):
        (
            self.config.runner.pcmk.load_state_plaintext(
                stdout="crm_mon cluster status",
            )
            .fs.exists(settings.corosync_conf_file, return_value=True)
            .corosync_conf.load()
            .runner.cib.load(
                resources="""
                <resources>
                    <primitive id="S1" class="stonith" type="fence_dummy">
                        <instance_attributes>
                            <nvpair name="method" value="cycle" />
                        </instance_attributes>
                    </primitive>
                    <primitive id="S2" class="stonith" type="fence_dummy">
                        <instance_attributes>
                            <nvpair name="action" value="value" />
                            <nvpair name="method" value="not-cycle" />
                        </instance_attributes>
                    </primitive>
                    <primitive id="S3" class="stonith" type="fence_dummy">
                        <instance_attributes>
                            <nvpair name="method" value="cycle" />
                            <nvpair name="action" value="value" />
                        </instance_attributes>
                    </primitive>
                </resources>
            """
            )
            .services.is_running(
                "sbd", return_value=False, name="services.is_running.sbd"
            )
        )
        self._fixture_config_local_daemons()

        self.assertEqual(
            # pylint: disable=line-too-long
            status.full_cluster_status_plaintext(self.env_assist.get_env()),
            dedent(
                """\
                Cluster name: test99

                WARNINGS:
                Following stonith devices have the 'action' option set, it is recommended to set 'pcmk_off_action', 'pcmk_reboot_action' instead: 'S2', 'S3'
                Following stonith devices have the 'method' option set to 'cycle' which is potentially dangerous, please consider using 'onoff': 'S1', 'S3'

                crm_mon cluster status

                Daemon Status:
                  corosync: active/enabled
                  pacemaker: active/enabled
                  pcsd: active/enabled"""
            ),
        )

    def test_pcsd_status_issues(self):
        self.node_name_list = ["node1", "node2", "node3", "node4", "node5"]

        (
            self.config.env.set_known_nodes(self.node_name_list[1:])
            .runner.pcmk.can_fence_history_status(stderr="not supported")
            .runner.pcmk.load_state_plaintext(
                verbose=True,
                stdout="crm_mon cluster status",
            )
            .fs.exists(settings.corosync_conf_file, return_value=True)
            .corosync_conf.load(node_name_list=self.node_name_list)
            .runner.cib.load(
                resources="""
                <resources>
                    <primitive id="S" class="stonith" type="fence_dummy" />
                </resources>
            """
            )
            .runner.pcmk.load_ticket_state_plaintext(stdout="ticket status")
            .services.is_running(
                "sbd", return_value=False, name="services.is_running.sbd"
            )
        )
        self._fixture_config_local_daemons()
        (
            self.config.http.host.check_reachability(
                communication_list=[
                    # node1 has no record in known-hosts
                    dict(
                        label="node2",
                        was_connected=False,
                        errno=7,
                        error_msg="node2 error",
                    ),
                    dict(
                        label="node3",
                        response_code=401,
                        output="node3 output",
                    ),
                    dict(
                        label="node4",
                        response_code=500,
                        output="node4 output",
                    ),
                    dict(
                        label="node5",
                    ),
                ]
            )
        )

        self.assertEqual(
            status.full_cluster_status_plaintext(
                self.env_assist.get_env(), verbose=True
            ),
            dedent(
                """\
                Cluster name: test99
                crm_mon cluster status

                Tickets:
                  ticket status

                PCSD Status:
                  node1: Unable to authenticate
                  node2: Offline
                  node3: Unable to authenticate
                  node4: Online
                  node5: Online

                Daemon Status:
                  corosync: active/enabled
                  pacemaker: active/enabled
                  pcsd: active/enabled"""
            ),
        )

    def test_daemon_status_all_on(self):
        self._fixture_config_live_minimal()
        self._fixture_config_local_daemons(
            corosync_enabled=True,
            corosync_active=True,
            pacemaker_enabled=True,
            pacemaker_active=True,
            pacemaker_remote_enabled=True,
            pacemaker_remote_active=True,
            pcsd_enabled=True,
            pcsd_active=True,
            sbd_enabled=True,
            sbd_active=True,
        )
        self.assertEqual(
            status.full_cluster_status_plaintext(self.env_assist.get_env()),
            dedent(
                """\
                Cluster name: test99
                crm_mon cluster status

                Daemon Status:
                  corosync: active/enabled
                  pacemaker: active/enabled
                  pacemaker_remote: active/enabled
                  pcsd: active/enabled
                  sbd: active/enabled"""
            ),
        )

    def test_daemon_status_all_off(self):
        self._fixture_config_live_minimal()
        self._fixture_config_local_daemons(
            corosync_enabled=False,
            corosync_active=False,
            pacemaker_enabled=False,
            pacemaker_active=False,
            pacemaker_remote_enabled=False,
            pacemaker_remote_active=False,
            pcsd_enabled=False,
            pcsd_active=False,
            sbd_enabled=False,
            sbd_active=False,
        )
        self.assertEqual(
            status.full_cluster_status_plaintext(self.env_assist.get_env()),
            dedent(
                """\
                Cluster name: test99
                crm_mon cluster status

                Daemon Status:
                  corosync: inactive/disabled
                  pacemaker: inactive/disabled
                  pcsd: inactive/disabled"""
            ),
        )



from typing import (
    Any,
    Dict,
    Iterable,
    NewType,
    Type,
    TypeVar,
    Union,
)
from dataclasses import asdict, fields, is_dataclass

import dacite

from pcs.common import types


PrimitiveType = Union[str, int, float, bool, None]
DtoPayload = Dict[str, "SerializableType"]  # type: ignore
SerializableType = Union[  # type: ignore
    PrimitiveType,
    DtoPayload,  # type: ignore
    Iterable["SerializableType"],  # type: ignore
]

T = TypeVar("T")

ToDictMetaKey = NewType("ToDictMetaKey", str)
META_NAME = ToDictMetaKey("META_NAME")


class DataTransferObject:
    pass


def meta(name: str) -> Dict[str, str]:
    metadata: Dict[str, str] = dict()
    if name:
        metadata[META_NAME] = name
    return metadata


def _is_compatible_type(_type: Type, arg_index: int) -> bool:
    return (
        hasattr(_type, "__args__")
        and len(_type.__args__) >= arg_index
        and is_dataclass(_type.__args__[arg_index])
    )


def _convert_dict(
    klass: Type[DataTransferObject], obj_dict: DtoPayload
) -> DtoPayload:
    new_dict = dict()
    for _field in fields(klass):
        value = obj_dict[_field.name]
        if is_dataclass(_field.type):
            value = _convert_dict(_field.type, value)
        elif isinstance(value, list) and _is_compatible_type(_field.type, 0):
            value = [
                _convert_dict(_field.type.__args__[0], item) for item in value
            ]
        elif isinstance(value, dict) and _is_compatible_type(_field.type, 1):
            value = {
                item_key: _convert_dict(_field.type.__args__[1], item_val)
                for item_key, item_val in value.items()
            }
        new_dict[_field.metadata.get(META_NAME, _field.name)] = value
    return new_dict


def to_dict(obj: DataTransferObject) -> DtoPayload:
    return _convert_dict(obj.__class__, asdict(obj))


DtoType = TypeVar("DtoType", bound=DataTransferObject)


def _convert_payload(klass: Type[DtoType], data: DtoPayload) -> DtoPayload:
    new_dict = dict()
    for _field in fields(klass):
        value = data[_field.metadata.get(META_NAME, _field.name)]
        if is_dataclass(_field.type):
            value = _convert_payload(_field.type, value)
        elif isinstance(value, list) and _is_compatible_type(_field.type, 0):
            value = [
                _convert_payload(_field.type.__args__[0], item)
                for item in value
            ]
        elif isinstance(value, dict) and _is_compatible_type(_field.type, 1):
            value = {
                item_key: _convert_payload(_field.type.__args__[1], item_val)
                for item_key, item_val in value.items()
            }
        new_dict[_field.name] = value
    return new_dict


def from_dict(cls: Type[DtoType], data: DtoPayload) -> DtoType:
    return dacite.from_dict(
        data_class=cls,
        data=_convert_payload(cls, data),
        # NOTE: all enum types has to be listed here in key cast
        # see: https://github.com/konradhalas/dacite#casting
        config=dacite.Config(
            cast=[
                types.CibNvsetType,
                types.CibRuleInEffectStatus,
                types.CibRuleExpressionType,
                types.CorosyncTransportType,
                types.DrRole,
                types.ResourceRelationType,
            ]
        ),
    )


class ImplementsToDto:
    def to_dto(self) -> Any:
        raise NotImplementedError()


class ImplementsFromDto:
    @classmethod
    def from_dto(cls: Type[T], dto_obj: Any) -> T:
        raise NotImplementedError()



from pcs import settings
from pcs.common import (
    file_type_codes,
    reports,
)
from pcs.common.file import RawFileError
from pcs.common.tools import format_environment_error
from pcs.common.reports.item import ReportItem
from pcs.lib.communication.nodes import SendPcsdSslCertAndKey
from pcs.lib.communication.tools import run_and_raise
from pcs.lib.env import LibraryEnvironment
from pcs.lib.errors import LibraryError
from pcs.lib.node import get_existing_nodes_names


def synchronize_ssl_certificate(env: LibraryEnvironment, skip_offline=False):
    """
    Send the local pcsd SSL cert and key to all full nodes in the local cluster.

    Consider the pcs Web UI is accessed via an IP running as a resource in the
    cluster. When the IP is moved, the user's browser connects to the new node
    and we want it to get the same certificate to make the transition a
    seamless experience (otherwise the browser display a warning that the
    certificate has changed).
    Using pcsd Web UI on remote and guest nodes is not supported (pcs/pcsd
    depends on the corosanc.conf file being present on the local node) so we
    send the cert only to corossync (== full stack) nodes.
    """
    report_processor = env.report_processor
    target_factory = env.get_node_target_factory()
    cluster_nodes_names, report_list = get_existing_nodes_names(
        env.get_corosync_conf()
    )
    if not cluster_nodes_names:
        report_list.append(
            ReportItem.error(reports.messages.CorosyncConfigNoNodesDefined())
        )
    report_processor.report_list(report_list)

    try:
        with open(settings.pcsd_cert_location, "r") as file:
            ssl_cert = file.read()
    except EnvironmentError as e:
        report_processor.report(
            ReportItem.error(
                reports.messages.FileIoError(
                    file_type_codes.PCSD_SSL_CERT,
                    RawFileError.ACTION_READ,
                    format_environment_error(e),
                    file_path=settings.pcsd_cert_location,
                )
            )
        )
    try:
        with open(settings.pcsd_key_location, "r") as file:
            ssl_key = file.read()
    except EnvironmentError as e:
        report_processor.report(
            ReportItem.error(
                reports.messages.FileIoError(
                    file_type_codes.PCSD_SSL_KEY,
                    RawFileError.ACTION_READ,
                    format_environment_error(e),
                    file_path=settings.pcsd_key_location,
                )
            )
        )

    (
        target_report_list,
        target_list,
    ) = target_factory.get_target_list_with_reports(
        cluster_nodes_names, skip_non_existing=skip_offline
    )
    report_processor.report_list(target_report_list)

    if report_processor.has_errors:
        raise LibraryError()

    env.report_processor.report(
        ReportItem.info(
            reports.messages.PcsdSslCertAndKeyDistributionStarted(
                sorted([target.label for target in target_list])
            )
        )
    )

    com_cmd = SendPcsdSslCertAndKey(env.report_processor, ssl_cert, ssl_key)
    com_cmd.set_targets(target_list)
    run_and_raise(env.get_node_communicator(), com_cmd)



from tornado.web import (
    RequestHandler,
    RedirectHandler as TornadoRedirectHandler,
)


class EnhanceHeadersMixin:
    """
    EnhanceHeadersMixin allows to add security headers to GUI urls.
    """

    def set_strict_transport_security(self):
        # rhbz 1558063
        # The HTTP Strict-Transport-Security response header (often abbreviated
        # as HSTS)  lets a web site tell browsers that it should only be
        # accessed using HTTPS, instead of using HTTP.
        self.set_header("Strict-Transport-Security", "max-age=604800")

    def set_header_nosniff_content_type(self):
        # The X-Content-Type-Options response HTTP header is a marker used by
        # the server to indicate that the MIME types advertised in the
        # Content-Type headers should not be changed and be followed. This
        # allows to opt-out of MIME type sniffing, or, in other words, it is a
        # way to say that the webmasters knew what they were doing.
        self.set_header("X-Content-Type-Options", "nosniff")

    def enhance_headers(self):
        self.set_header_nosniff_content_type()

        # The X-Frame-Options HTTP response header can be used to indicate
        # whether or not a browser should be allowed to render a page in a
        # <frame>, <iframe> or <object> . Sites can use this to avoid
        # clickjacking attacks, by ensuring that their content is not embedded
        # into other sites.
        self.set_header("X-Frame-Options", "SAMEORIGIN")

        # The HTTP Content-Security-Policy (CSP) frame-ancestors directive
        # specifies valid parents that may embed a page using <frame>,
        # <iframe>, <object>, <embed>, or <applet>.

        # Requested upstream:
        # https://lists.clusterlabs.org/pipermail/users/2020-May/027199.html

        # For now, I'm just setting this to the same value as already present
        # X-Frame-Options header to keep them consistent.
        self.set_header("Content-Security-Policy", "frame-ancestors 'self'")

        # The HTTP X-XSS-Protection response header is a feature of Internet
        # Explorer, Chrome and Safari that stops pages from loading when they
        # detect reflected cross-site scripting (XSS) attacks. Although these
        # protections are largely unnecessary in modern browsers when sites
        # implement a strong Content-Security-Policy that disables the use of
        # inline JavaScript ('unsafe-inline'), they can still provide
        # protections for users of older web browsers that don't yet support
        # CSP.
        self.set_header("X-Xss-Protection", "1; mode=block")


class BaseHandler(EnhanceHeadersMixin, RequestHandler):
    """
    BaseHandler adds for all urls Strict-Transport-Security.
    """

    def set_default_headers(self):
        self.set_strict_transport_security()

    def data_received(self, chunk):
        # abstract method `data_received` does need to be overriden. This
        # method should be implemented to handle streamed request data.
        # BUT we currently do not plan to use it SO:
        # pylint: disable=abstract-method
        pass


class RedirectHandler(EnhanceHeadersMixin, TornadoRedirectHandler):
    """
    RedirectHandler with Strict-Transport-Security headers.
    """

    # abstract method `data_received` is not used in redirect:
    # pylint: disable=abstract-method
    def initialize(self, url, permanent=True):
        super().initialize(url, permanent)
        self.set_strict_transport_security()



from unittest import TestCase

from pcs_test.tier0.lib.commands.resource.bundle_common import (
    FixturesMixin,
    SetUpMixin,
    ParametrizedContainerMixin,
    NetworkMixin,
    PortMapMixin,
    StorageMapMixin,
    MetaMixin,
    AllOptionsMixin,
    WaitMixin,
)
from pcs_test.tools import fixture

from pcs.common.reports import ReportItemSeverity as severities
from pcs.common.reports import codes as report_codes
from pcs.lib.commands.resource import bundle_reset


class BaseMixin(FixturesMixin):
    container_type = None
    bundle_id = "B1"
    image = "pcs:test"

    @property
    def initial_resources(self):
        return self.fixture_resources_bundle_simple

    def bundle_reset(self, bundle_id=None, **params):
        if "container_options" not in params:
            params["container_options"] = {"image": self.image}

        bundle_reset(
            self.env_assist.get_env(),
            bundle_id=bundle_id or self.bundle_id,
            **params,
        )

    def run_bundle_cmd(self, *args, **kwargs):
        self.bundle_reset(*args, **kwargs)


class MinimalMixin(BaseMixin, SetUpMixin):
    container_type = None
    initial_cib_filename = "cib-empty-3.2.xml"

    def test_success_zero_change(self):
        # Resets a bundle with only an image set to a bundle with the same
        # image set and no other options.
        self.config.env.push_cib(resources=self.initial_resources)
        self.bundle_reset()

    def test_success_change(self):
        new_image = "{0}:new".format(self.image)

        self.config.env.push_cib(
            replace={
                ".//resources/bundle": """
                    <bundle id="{bundle_id}">
                        <{container_type} image="{image}" />
                    </bundle>
                """.format(
                    bundle_id=self.bundle_id,
                    container_type=self.container_type,
                    image=new_image,
                ),
            }
        )
        self.bundle_reset(
            container_options={"image": new_image},
        )

    def test_noexistent_id(self):
        self.env_assist.assert_raise_library_error(
            lambda: self.bundle_reset(bundle_id="B0"),
            [
                (
                    severities.ERROR,
                    report_codes.ID_NOT_FOUND,
                    {
                        "id": "B0",
                        "expected_types": ["bundle"],
                        "context_type": "resources",
                        "context_id": "",
                    },
                    None,
                ),
            ],
            expected_in_processor=False,
        )

    def test_no_options_set(self):
        self.env_assist.assert_raise_library_error(
            lambda: bundle_reset(self.env_assist.get_env(), self.bundle_id)
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.REQUIRED_OPTIONS_ARE_MISSING,
                    option_names=["image"],
                    option_type="container",
                ),
            ]
        )


class FullMixin(SetUpMixin, BaseMixin):
    container_type = None
    fixture_primitive = """
        <primitive class="ocf" id="A" provider="heartbeat" type="Dummy"/>
    """

    @property
    def initial_resources(self):
        return """
            <resources>
                <bundle id="{bundle_id}">
                    <{container_type}
                        image="{image}"
                        promoted-max="0"
                        replicas="1"
                        replicas-per-host="1"
                    />
                    <network
                        control-port="12345"
                        host-interface="eth0"
                        host-netmask="24"
                        ip-range-start="192.168.100.200"
                    >
                        <port-mapping
                            id="{bundle_id}-port-map-1001"
                            internal-port="2002"
                            port="2000"
                        />
                        <port-mapping
                            id="{bundle_id}-port-map-3000-3300"
                            range="3000-3300"
                        />
                    </network>
                    <storage>
                        <storage-mapping
                            id="{bundle_id}-storage-map"
                            options="extra options 2"
                            source-dir="/tmp/{container_type}2a"
                            target-dir="/tmp/{container_type}2b"
                        />
                    </storage>
                    <meta_attributes id="{bundle_id}-meta_attributes">
                        <nvpair
                            id="{bundle_id}-meta_attributes-target-role"
                            name="target-role"
                            value="Stopped"
                        />
                    </meta_attributes>
                    {fixture_primitive}
                </bundle>
            </resources>
        """.format(
            container_type=self.container_type,
            bundle_id=self.bundle_id,
            fixture_primitive=self.fixture_primitive,
            image=self.image,
        )

    def test_success_minimal(self):
        new_image = "{0}:new".format(self.image)

        # Garbage (empty tags network, storage and meta_attributes) are kept.
        # See https://bugzilla.redhat.com/show_bug.cgi?id=1642514
        self.config.env.push_cib(
            replace={
                ".//resources/bundle": """
                    <bundle id="{bundle_id}">
                        <{container_type} image="{image}" />
                        <network/>
                        <storage/>
                        <meta_attributes id="{bundle_id}-meta_attributes"/>
                        {fixture_primitive}
                    </bundle>
                """.format(
                    container_type=self.container_type,
                    bundle_id=self.bundle_id,
                    fixture_primitive=self.fixture_primitive,
                    image=new_image,
                ),
            }
        )

        self.bundle_reset(
            container_options={"image": new_image},
        )

    def test_success_full(self):
        new_image = "{0}:new".format(self.image)

        self.config.env.push_cib(
            replace={
                ".//resources/bundle": """
                    <bundle id="{bundle_id}">
                        <{container_type}
                            image="{image}"
                            promoted-max="1"
                            replicas="2"
                            replicas-per-host="3"
                        />
                        <network
                            control-port="54321"
                            host-interface="eth1"
                            host-netmask="16"
                            ip-range-start="192.168.100.1"
                        >
                            <port-mapping
                                id="{bundle_id}-port-map-3000"
                                internal-port="3002"
                                port="3000"
                            />
                            <port-mapping
                                id="{bundle_id}-port-map-4000-4400"
                                range="4000-4400"
                            />
                        </network>
                        <storage>
                            <storage-mapping
                                id="{bundle_id}-storage-map"
                                options="extra options 2"
                                source-dir="/tmp/{container_type}2aa"
                                target-dir="/tmp/{container_type}2bb"
                            />
                        </storage>
                        <meta_attributes id="{bundle_id}-meta_attributes">
                            <nvpair
                                id="{bundle_id}-meta_attributes-target-role"
                                name="target-role"
                                value="Started"
                            />
                        </meta_attributes>
                        {fixture_primitive}
                    </bundle>
                """.format(
                    container_type=self.container_type,
                    bundle_id=self.bundle_id,
                    fixture_primitive=self.fixture_primitive,
                    image=new_image,
                ),
            }
        )
        self.bundle_reset(
            container_options={
                "image": new_image,
                "promoted-max": "1",
                "replicas": "2",
                "replicas-per-host": "3",
            },
            network_options={
                "control-port": "54321",
                "host-interface": "eth1",
                "host-netmask": "16",
                "ip-range-start": "192.168.100.1",
            },
            port_map=[
                {"internal-port": "3002", "port": "3000"},
                {"range": "4000-4400"},
            ],
            storage_map=[
                {
                    "options": "extra options 2",
                    "source-dir": f"/tmp/{self.container_type}2aa",
                    "target-dir": f"/tmp/{self.container_type}2bb",
                },
            ],
            meta_attributes={
                "target-role": "Started",
            },
        )

    def test_success_keep_map_ids(self):
        self.config.env.push_cib(
            replace={
                ".//resources/bundle/network": f"""
                    <network
                        control-port="12345"
                        host-interface="eth0"
                        host-netmask="24"
                        ip-range-start="192.168.100.200"
                    >
                        <port-mapping
                            id="{self.bundle_id}-port-map-1001"
                            internal-port="3002"
                            port="3000"
                        />
                        <port-mapping
                            id="{self.bundle_id}-port-map-3000-3300"
                            range="4000-4400"
                        />
                    </network>
                """,
                ".//resources/bundle/storage": f"""
                    <storage>
                        <storage-mapping
                            id="{self.bundle_id}-storage-map"
                            options="extra options 2"
                            source-dir="/tmp/{self.container_type}2aa"
                            target-dir="/tmp/{self.container_type}2bb"
                        />
                    </storage>
                """,
            }
        )

        # Every value is kept as before except port_map and storage_map.
        self.bundle_reset(
            container_options={
                "image": self.image,
                "promoted-max": "0",
                "replicas": "1",
                "replicas-per-host": "1",
            },
            network_options={
                "control-port": "12345",
                "host-interface": "eth0",
                "host-netmask": "24",
                "ip-range-start": "192.168.100.200",
            },
            port_map=[
                {
                    "id": f"{self.bundle_id}-port-map-1001",
                    "internal-port": "3002",
                    "port": "3000",
                },
                {
                    "id": f"{self.bundle_id}-port-map-3000-3300",
                    "range": "4000-4400",
                },
            ],
            storage_map=[
                {
                    "id": f"{self.bundle_id}-storage-map",
                    "options": "extra options 2",
                    "source-dir": f"/tmp/{self.container_type}2aa",
                    "target-dir": f"/tmp/{self.container_type}2bb",
                },
            ],
            meta_attributes={
                "target-role": "Stopped",
            },
        )


class ResetParametrizedContainerMixin(BaseMixin, ParametrizedContainerMixin):
    pass


class MinimalRkt(MinimalMixin, TestCase):
    container_type = "rkt"


class MinimalPodman(MinimalMixin, TestCase):
    container_type = "podman"


class MinimalDocker(MinimalMixin, TestCase):
    container_type = "docker"


class FullRkt(FullMixin, TestCase):
    container_type = "rkt"


class FullPodman(FullMixin, TestCase):
    container_type = "podman"


class FullDocker(FullMixin, TestCase):
    container_type = "docker"


class ResetParametrizedPodman(ResetParametrizedContainerMixin, TestCase):
    container_type = "podman"


class ResetParametrizedDocker(ResetParametrizedContainerMixin, TestCase):
    container_type = "docker"


class ResetParametrizedRkt(ResetParametrizedContainerMixin, TestCase):
    container_type = "rkt"


class ResetWithNetwork(BaseMixin, NetworkMixin, TestCase):
    container_type = "docker"


class ResetWithPortMap(BaseMixin, PortMapMixin, TestCase):
    container_type = "docker"


class ResetWithStorageMap(BaseMixin, StorageMapMixin, TestCase):
    container_type = "docker"


class ResetWithMetaMap(BaseMixin, MetaMixin, TestCase):
    container_type = "docker"

    def test_success(self):
        # When there is no meta attributes the new one are put on the first
        # possition (since reset now uses update internally). This is the reason
        # for reimplementation of this MetaMixin test.
        self.config.env.push_cib(
            resources="""
                <resources>
                    <bundle id="{bundle_id}">
                        <meta_attributes id="{bundle_id}-meta_attributes">
                            <nvpair id="{bundle_id}-meta_attributes-is-managed"
                                name="is-managed" value="false" />
                            <nvpair id="{bundle_id}-meta_attributes-target-role"
                                name="target-role" value="Stopped" />
                        </meta_attributes>
                        <{container_type} image="{image}" />
                    </bundle>
                </resources>
            """.format(
                container_type=self.container_type,
                bundle_id=self.bundle_id,
                image=self.image,
            )
        )
        self.run_bundle_cmd(
            meta_attributes={
                "target-role": "Stopped",
                "is-managed": "false",
            }
        )


class ResetWithAllOptions(BaseMixin, AllOptionsMixin, TestCase):
    container_type = "docker"


class ResetWithWait(BaseMixin, WaitMixin, TestCase):
    container_type = "docker"


class ResetUnknownContainerType(BaseMixin, SetUpMixin, TestCase):
    container_type = "unknown"

    def test_error_or_unknown_container(self):
        self.env_assist.assert_raise_library_error(
            lambda: bundle_reset(self.env_assist.get_env(), self.bundle_id)
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.RESOURCE_BUNDLE_UNSUPPORTED_CONTAINER_TYPE,
                    bundle_id="B1",
                    supported_container_types=["docker", "podman", "rkt"],
                ),
            ]
        )


class NoMetaIdRegenerationMixin(BaseMixin, SetUpMixin):
    @property
    def initial_resources(self):
        return """
            <resources>
                <bundle id="{bundle_id}">
                    <{container_type}
                        image="{image}"
                        promoted-max="0"
                        replicas="1"
                        replicas-per-host="1"
                    />
                    <meta_attributes id="CUSTOM_ID">
                        <nvpair
                            id="ANOTHER_ID-target-role"
                            name="target-role"
                            value="Stopped"
                        />
                    </meta_attributes>
                </bundle>
            </resources>
        """.format(
            container_type=self.container_type,
            bundle_id=self.bundle_id,
            image=self.image,
        )

    def test_dont_regenerate_meta_attributes_id(self):
        self.config.env.push_cib(
            replace={
                ".//resources/bundle/meta_attributes": """
                    <meta_attributes id="CUSTOM_ID">
                        <nvpair
                            id="CUSTOM_ID-target-role"
                            name="target-role"
                            value="Stopped"
                        />
                    </meta_attributes>
                """,
            }
        )
        self.bundle_reset(
            container_options={
                "image": self.image,
                "promoted-max": "0",
                "replicas": "1",
                "replicas-per-host": "1",
            },
            meta_attributes={
                "target-role": "Stopped",
            },
        )


class NoMetaIdRegenerationDocker(NoMetaIdRegenerationMixin, TestCase):
    container_type = "docker"


class NoMetaIdRegenerationPodman(NoMetaIdRegenerationMixin, TestCase):
    container_type = "podman"


class NoMetaIdRegenerationRkt(NoMetaIdRegenerationMixin, TestCase):
    container_type = "rkt"



from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.parse_args import group_by_keywords, prepare_options


def config_setup(lib, arg_list, modifiers):
    """
    create booth config

    Options:
      * --force - overwrite existing
      * --booth-conf - booth config file
      * --booth-key - booth authkey file
      * --name - name of a booth instance
    """
    modifiers.ensure_only_supported(
        "--force",
        "--booth-conf",
        "--booth-key",
        "--name",
    )
    peers = group_by_keywords(
        arg_list, set(["sites", "arbitrators"]), keyword_repeat_allowed=False
    )
    if "sites" not in peers or not peers["sites"]:
        raise CmdLineInputError()

    lib.booth.config_setup(
        peers["sites"],
        peers["arbitrators"],
        instance_name=modifiers.get("--name"),
        overwrite_existing=modifiers.get("--force"),
    )


def config_destroy(lib, arg_list, modifiers):
    """
    destroy booth config

    Options:
      --force - ignore config load issues
      --name - name of a booth instance
    """
    modifiers.ensure_only_supported("--force", "--name")
    if arg_list:
        raise CmdLineInputError()
    lib.booth.config_destroy(
        instance_name=modifiers.get("--name"),
        ignore_config_load_problems=modifiers.get("--force"),
    )


def config_show(lib, arg_list, modifiers):
    """
    print booth config

    Options:
      * --name - name of a booth instace
      * --request-timeout - HTTP timeout for getting config from remote host
    """
    modifiers.ensure_only_supported("--name", "--request-timeout")
    if len(arg_list) > 1:
        raise CmdLineInputError()
    node = None if not arg_list else arg_list[0]

    print(
        lib.booth.config_text(
            instance_name=modifiers.get("--name"), node_name=node
        )
        .decode("utf-8")
        .rstrip()
    )


def config_ticket_add(lib, arg_list, modifiers):
    """
    add ticket to current configuration

    Options:
      * --force
      * --booth-conf - booth config file
      * --booth-key - booth auth key file
      * --name - name of a booth instace
    """
    modifiers.ensure_only_supported(
        "--force", "--booth-conf", "--name", "--booth-key"
    )
    if not arg_list:
        raise CmdLineInputError
    lib.booth.config_ticket_add(
        arg_list[0],
        prepare_options(arg_list[1:]),
        instance_name=modifiers.get("--name"),
        allow_unknown_options=modifiers.get("--force"),
    )


def config_ticket_remove(lib, arg_list, modifiers):
    """
    add ticket to current configuration

    Options:
      * --booth-conf - booth config file
      * --booth-key - booth auth key file
      * --name - name of a booth instace
    """
    modifiers.ensure_only_supported("--booth-conf", "--name", "--booth-key")
    if len(arg_list) != 1:
        raise CmdLineInputError
    lib.booth.config_ticket_remove(
        arg_list[0],
        instance_name=modifiers.get("--name"),
    )


def _ticket_operation(lib_call, arg_list, booth_name):
    """
    Commandline options:
      * --name - name of a booth instance
    """
    site_ip = None
    if len(arg_list) == 2:
        site_ip = arg_list[1]
    elif len(arg_list) != 1:
        raise CmdLineInputError()

    ticket = arg_list[0]
    lib_call(ticket, site_ip=site_ip, instance_name=booth_name)


def ticket_revoke(lib, arg_list, modifiers):
    """
    Options:
      * --name - name of a booth instance
    """
    modifiers.ensure_only_supported("--name")
    _ticket_operation(
        lib.booth.ticket_revoke, arg_list, modifiers.get("--name")
    )


def ticket_grant(lib, arg_list, modifiers):
    """
    Options:
      * --name - name of a booth instance
    """
    modifiers.ensure_only_supported("--name")
    _ticket_operation(lib.booth.ticket_grant, arg_list, modifiers.get("--name"))


def create_in_cluster(lib, arg_list, modifiers):
    """
    Options:
      * --force - allows to create booth resource even if its agent is not
        installed
      * -f - CIB file
      * --name - name of a booth instance
    """
    modifiers.ensure_only_supported("--force", "-f", "--name")
    if len(arg_list) != 2 or arg_list[0] != "ip":
        raise CmdLineInputError()
    lib.booth.create_in_cluster(
        arg_list[1],
        instance_name=modifiers.get("--name"),
        allow_absent_resource_agent=modifiers.get("--force"),
    )


def get_remove_from_cluster(resource_remove):
    # TODO resource_remove is provisional hack until resources are not moved to
    # lib
    def remove_from_cluster(lib, arg_list, modifiers):
        """
        Options:
          * --force - allow remove of multiple
          * -f - CIB file
          * --name - name of a booth instance
        """
        modifiers.ensure_only_supported("--force", "-f", "--name")
        if arg_list:
            raise CmdLineInputError()

        lib.booth.remove_from_cluster(
            resource_remove,
            instance_name=modifiers.get("--name"),
            allow_remove_multiple=modifiers.get("--force"),
        )

    return remove_from_cluster


def get_restart(resource_restart):
    # TODO resource_restart is provisional hack until resources are not moved to
    # lib
    def restart(lib, arg_list, modifiers):
        """
        Options:
          * --force - allow multiple
          * --name - name of a booth instance
        """
        modifiers.ensure_only_supported("--force", "--name")
        if arg_list:
            raise CmdLineInputError()

        lib.booth.restart(
            lambda resource_id_list: resource_restart(
                lib, resource_id_list, modifiers.get_subset("--force")
            ),
            instance_name=modifiers.get("--name"),
            allow_multiple=modifiers.get("--force"),
        )

    return restart


def sync(lib, arg_list, modifiers):
    """
    Options:
      * --skip-offline - skip offline nodes
      * --name - name of a booth instance
      * --booth-conf - booth config file
      * --booth-key - booth authkey file
      * --request-timeout - HTTP timeout for file ditribution
    """
    modifiers.ensure_only_supported(
        "--skip-offline",
        "--name",
        "--booth-conf",
        "--booth-key",
        "--request-timeout",
    )
    if arg_list:
        raise CmdLineInputError()
    lib.booth.config_sync(
        instance_name=modifiers.get("--name"),
        skip_offline_nodes=modifiers.get("--skip-offline"),
    )


def enable(lib, arg_list, modifiers):
    """
    Options:
      * --name - name of a booth instance
    """
    modifiers.ensure_only_supported("--name")
    if arg_list:
        raise CmdLineInputError()
    lib.booth.enable_booth(instance_name=modifiers.get("--name"))


def disable(lib, arg_list, modifiers):
    """
    Options:
      * --name - name of a booth instance
    """
    modifiers.ensure_only_supported("--name")
    if arg_list:
        raise CmdLineInputError()
    lib.booth.disable_booth(instance_name=modifiers.get("--name"))


def start(lib, arg_list, modifiers):
    """
    Options:
      * --name - name of a booth instance
    """
    modifiers.ensure_only_supported("--name")
    if arg_list:
        raise CmdLineInputError()
    lib.booth.start_booth(instance_name=modifiers.get("--name"))


def stop(lib, arg_list, modifiers):
    """
    Options:
      * --name - name of a booth instance
    """
    modifiers.ensure_only_supported("--name")
    if arg_list:
        raise CmdLineInputError()
    lib.booth.stop_booth(instance_name=modifiers.get("--name"))


def pull(lib, arg_list, modifiers):
    """
    Options:
      * --name - name of a booth instance
      * --request-timeout - HTTP timeout for file ditribution
    """
    modifiers.ensure_only_supported("--name", "--request-timeout")
    if len(arg_list) != 1:
        raise CmdLineInputError()
    lib.booth.pull_config(
        arg_list[0],
        instance_name=modifiers.get("--name"),
    )


def status(lib, arg_list, modifiers):
    """
    Options:
      * --name - name of booth instance
    """
    modifiers.ensure_only_supported("--name")
    if arg_list:
        raise CmdLineInputError()
    booth_status = lib.booth.get_status(instance_name=modifiers.get("--name"))
    if booth_status.get("ticket"):
        print("TICKETS:")
        print(booth_status["ticket"])
    if booth_status.get("peers"):
        print("PEERS:")
        print(booth_status["peers"])
    if booth_status.get("status"):
        print("DAEMON STATUS:")
        print(booth_status["status"])
